import logging

from ._filter import Filter
from ._logger import Logger
from .formatters import Formatter, SimpleFormatter

__all__ = ("Logger", "Filter", "Formatter", "SimpleFormatter")


# Set custom Logger class

_old_logger_class = logging.getLoggerClass()
logging.setLoggerClass(Logger)

# Register Loggers

logger: Logger = logging.getLogger("jj.access_logger")  # type: ignore
logger.setLevel(logging.INFO)
logger.propagate = False

# Register Filter

filter_ = Filter()
logger.addFilter(filter_)

# Register Handler

handler = logging.StreamHandler()
logger.addHandler(handler)

# Register Formatter

formatter = SimpleFormatter()
handler.setFormatter(formatter)

# Restore default Logger class

logging.setLoggerClass(_old_logger_class)



from typing import Any, Dict, Union

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, DictOrTupleList, MultiDictMatcher
from ._request_matcher import RequestMatcher

__all__ = ("ParamMatcher", "DictOrTupleListOrAttrMatcher",)


DictOrTupleListOrAttrMatcher = Union[DictOrTupleList, AttributeMatcher]


@packable("jj.matchers.ParamMatcher")
class ParamMatcher(RequestMatcher):
    def __init__(self, params: DictOrTupleListOrAttrMatcher, *, resolver: Resolver) -> None:
        super().__init__(resolver=resolver)
        if isinstance(params, AttributeMatcher):
            self._matcher = params
        else:
            self._matcher = MultiDictMatcher(params)

    async def match(self, request: Request) -> bool:
        return await self._matcher.match(request.query)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"params": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     params: DictOrTupleListOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "ParamMatcher":
        return cls(params, resolver=resolver)



import unittest

from jj.apps import BaseApp, define_app


class TestDefineApp(unittest.TestCase):
    def test_define_app(self):
        self.assertTrue(issubclass(define_app(), BaseApp))

    def test_define_app_with_name(self):
        name = "CustomApp"
        app = define_app(name)
        self.assertEqual(app.__name__, name)



from typing import Any, Dict, Optional

from aiohttp import web
from multidict import MultiDictProxy

from ..responses import StreamResponse

__all__ = ("Request",)


class Request(web.Request):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._segments: Optional[Dict[str, str]] = None

    @property
    def params(self) -> "MultiDictProxy[str]":
        return self.query

    @property
    def segments(self) -> Dict[str, str]:
        if self._segments is None:
            return {}
        return self._segments

    @segments.setter
    def segments(self, segments: Optional[Dict[str, str]]) -> None:
        self._segments = segments

    async def _prepare_hook(self, response: StreamResponse) -> None:  # type: ignore
        pass



from typing import Any, Dict, List, Tuple

from aiohttp.web_exceptions import HTTPRequestEntityTooLarge
from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy
from packed import packable

from ...requests import Request

__all__ = ("HistoryRequest",)


@packable("jj.mock.HistoryRequest")
class HistoryRequest:
    def __init__(self, *,
                 method: str,
                 path: str,
                 segments: Dict[str, str],
                 params: "MultiDictProxy[str]",
                 headers: "CIMultiDictProxy[str]",
                 body: Any,
                 raw: bytes) -> None:
        self._method = method
        self._path = path
        self._segments = segments
        self._params = params
        self._headers = headers
        self._body = body
        self._raw = raw

    @property
    def method(self) -> str:
        return self._method

    @property
    def path(self) -> str:
        return self._path

    @property
    def segments(self) -> Dict[str, str]:
        return self._segments

    @property
    def params(self) -> "MultiDictProxy[str]":
        return self._params

    @property
    def headers(self) -> "CIMultiDictProxy[str]":
        return self._headers

    @property
    def body(self) -> Any:
        return self._body

    @property
    def raw(self) -> bytes:
        return self._raw

    @classmethod
    async def from_request(cls, request: Request) -> "HistoryRequest":
        try:
            raw = await request.read()
        except HTTPRequestEntityTooLarge:
            raw = b"<binary>"
            await request.release()
        return cls(
            method=request.method,
            path=request.path,
            segments=request.segments,
            params=request.params,
            headers=request.headers,
            body=raw,
            raw=raw,
        )

    def to_dict(self) -> Dict[str, Any]:
        params = [[key, val] for key, val in self._params.items()]
        headers = [[key, val] for key, val in self._headers.items()]
        return {
            "method": self._method,
            "path": self._path,
            "segments": self._segments,
            "params": params,
            "headers": headers,
            "body": self._body,
            "raw": self._raw,
        }

    def __packed__(self) -> Dict[str, Any]:
        return self.to_dict()

    @classmethod
    def from_dict(cls, request: Dict[str, Any]) -> "HistoryRequest":
        real_params = MultiDictProxy(MultiDict(request["params"]))
        real_headers = CIMultiDictProxy(CIMultiDict(request["headers"]))
        raw = request.get("raw", request["body"])  # backward compatibility
        return cls(
            method=request["method"],
            path=request["path"],
            segments=request["segments"],
            params=real_params,
            headers=real_headers,
            body=request["body"],
            raw=raw,
        )

    @classmethod
    def __unpacked__(cls, *,
                     method: str,
                     path: str,
                     segments: Dict[str, str],
                     params: List[Tuple[str, str]],
                     headers: List[Tuple[str, str]],
                     body: Any,
                     **kwargs: Any) -> "HistoryRequest":
        return cls.from_dict({
            "method": method,
            "path": path,
            "segments": segments,
            "params": params,
            "headers": headers,
            "body": body,
            **kwargs,
        })

    def __repr__(self) -> str:
        return (f"HistoryRequest("
                f"method={self._method!r}, "
                f"path={self._path!r}, "
                f"params={self._params!r}, "
                f"headers={self._headers!r}, "
                f"body={self._body!r}"
                f")")



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, sentinel

import pytest

from jj.requests import Request
from jj.resolvers import Resolver


@pytest.fixture
def resolver_():
    return Mock(Resolver)


@pytest.fixture
def request_():
    return Mock(Request)


@pytest.fixture
def handler_():
    return AsyncMock(return_value=sentinel.response)



from aiohttp.web import ContentCoding

from jj._version import server_version
from jj.responses import Response

from .._test_utils.steps import given, then, when


def test_unpack_binary():
    with given:
        packed = {
            "status": 201,
            "reason": None,
            "headers": [["Server", server_version]],
            "cookies": [],
            "body": b"201 Created",
            "chunked": False,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == {**packed, "reason": "Created"}


def test_unpack_text():
    with given:
        body = b"202 Accepted"
        packed = {
            "status": 202,
            "reason": "Accepted",
            "headers": [
                ["Server", server_version],
                ["Content-Length", str(len(body))],
                ["Content-Type", "text/plain; charset=utf-8"],
            ],
            "cookies": [],
            "body": body,
            "chunked": False,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == packed


def test_unpack_no_body():
    with given:
        packed = {
            "status": 200,
            "reason": "OK",
            "headers": [
                ["Server", server_version],
                ["Content-Length", "0"],
                ["Content-Type", "text/plain; charset=utf-8"],
            ],
            "cookies": [],
            "body": b"",
            "chunked": False,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == {**packed}


def test_unpack_cookies():
    with given:
        packed = {
            "status": 200,
            "reason": "OK",
            "headers": [["Server", server_version]],
            "cookies": [
                {
                    "name": "name1",
                    "value": "value1",
                    "expires": "Thu, 01 Jan 1970 00:00:01 GMT",
                    "domain": ".localhost",
                    "max_age": "0",
                    "path": "/path",
                    "secure": "secure",
                    "httponly": "httponly",
                    "version": "1",
                },
                {
                    "name": "name2",
                    "value": "value2",
                    "expires": None,
                    "domain": None,
                    "max_age": None,
                    "path": "/",
                    "secure": None,
                    "httponly": None,
                    "version": None,
                },
            ],
            "body": b"200 OK",
            "chunked": False,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == packed


def test_unpack_chunked():
    with given:
        packed = {
            "status": 200,
            "reason": "OK",
            "headers": [
                ["Server", server_version],
                ["X-Correlation-ID", "1234"],
            ],
            "cookies": [],
            "body": b"200 OK",
            "chunked": True,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == packed


def test_unpack_compressed():
    with given:
        packed = {
            "status": 200,
            "reason": "OK",
            "headers": [["Server", server_version]],
            "cookies": [],
            "body": b"200 OK",
            "chunked": False,
            "compression": ContentCoding.gzip,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == {**packed, "compression": ContentCoding.gzip.value}


def test_unpack_text_body_with_headers():
    with given:
        body = b"200 OK"
        packed = {
            "status": 200,
            "reason": "Accepted",
            "headers": [
                ["X-Custom-Header1", "Value1"],
                ["X-Custom-Header2", "Value2"],
                ["Server", server_version],
                ["Content-Length", str(len(body))],
                ["Content-Type", "text/plain; charset=utf-8"],
            ],
            "cookies": [],
            "body": body,
            "chunked": False,
            "compression": None,
        }

    with when:
        response = Response.__unpacked__(**packed)

    with then:
        assert response.__packed__() == packed



import mimetypes

from ._relay_response import RelayResponse
from ._response import Response
from ._static_response import StaticResponse
from ._stream_response import StreamResponse

__all__ = ("Response", "StaticResponse", "StreamResponse", "RelayResponse")


mimetypes.add_type("application/json", ".json")



from typing import Any, Dict, Optional
from urllib.parse import urljoin

from aiohttp import ClientSession
from aiohttp.abc import AbstractStreamWriter
from aiohttp.web_request import BaseRequest
from multidict import CIMultiDict, MultiDict
from packed import packable

from ._stream_response import StreamResponse

__all__ = ("RelayResponse",)


# http://tools.ietf.org/html/rfc2616#section-13.5.1
_HOP_BY_HOP_HEADERS = (
    "connection",
    "keep-alive",
    "proxy-authenticate",
    "proxy-authorization",
    "te",
    "trailers",
    "transfer-encoding",
    "upgrade")

_FILTERED_HEADERS = _HOP_BY_HOP_HEADERS + ("host", "content-length")


@packable("jj.responses.RelayResponse")
class RelayResponse(StreamResponse):
    def __init__(self, *, target: str) -> None:
        super().__init__()
        self._target = target

    @property
    def target(self) -> str:
        return self._target

    async def prepare(self, request: BaseRequest) -> Optional[AbstractStreamWriter]:
        url = urljoin(self.target, request.path)

        headers: MultiDict[str] = MultiDict()
        for key, value in request.headers.items():
            if key.lower() not in _FILTERED_HEADERS:
                headers[key] = value

        async with ClientSession(auto_decompress=False) as session:
            async with session.request(request.method, url,
                                       params=request.query,
                                       headers=headers,
                                       data=request.content,
                                       allow_redirects=True) as response:
                self.set_status(response.status, response.reason)
                self._headers = CIMultiDict(response.headers)

                await super().prepare(request)

                async for data in response.content.iter_any():
                    await self.write(data)

                await self.write_eof()

        return await super().prepare(request)

    def copy(self) -> "RelayResponse":
        assert not self.prepared
        return self.__class__(target=self._target)

    def __packed__(self) -> Dict[str, Any]:
        return {"target": self._target}

    @classmethod
    def __unpacked__(cls, *, target: str, **kwargs: Any) -> "RelayResponse":
        return cls(target=target)



import os
from distutils.util import strtobool
from typing import Optional, Union

from jj.expiration_policy import ExpirationPolicy
from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import (
    HistoryAdapterType,
    HistoryItem,
    HistoryRepository,
    HistoryRequest,
    HistoryResponse,
    default_history_adapter,
)
from ._mock import Mock
from ._mocked import Mocked
from ._remote_handler import RemoteHandler
from ._remote_mock import RemoteMock
from ._remote_response import RemoteResponseType
from ._system_log_filter import SystemLogFilter

_REMOTE_MOCK_URL = os.environ.get("JJ_REMOTE_MOCK_URL", "http://localhost:8080")
_REMOTE_MOCK_DISPOSABLE = os.environ.get("JJ_REMOTE_MOCK_DISPOSABLE", "True")


def mocked(matcher: Union[RequestMatcher, LogicalMatcher],
           response: RemoteResponseType,
           expiration_policy: Optional[ExpirationPolicy] = None,
           *,
           disposable: Optional[bool] = None,
           prefetch_history: bool = True,
           history_adapter: Optional[HistoryAdapterType] = default_history_adapter) -> "Mocked":
    if disposable is None:
        disposable = strtobool(_REMOTE_MOCK_DISPOSABLE)

    handler = RemoteMock(_REMOTE_MOCK_URL).create_handler(
        matcher,
        response,
        expiration_policy,
        history_adapter=history_adapter
    )
    return Mocked(handler, disposable=disposable, prefetch_history=prefetch_history)


__all__ = ("Mock", "mocked", "RemoteMock", "RemoteHandler", "Mocked",
           "HistoryRepository", "HistoryRequest", "HistoryResponse", "HistoryItem",
           "SystemLogFilter", "RemoteResponseType", "HistoryAdapterType",
           "default_history_adapter",)



import os
import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest

import jj
from jj import server_version
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import StaticResponse

from .._test_utils import run


class TestStaticResponse(TestCase):
    def make_app_with_response(self, *args, **kwargs):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return StaticResponse(*args, **kwargs)
        return App()

    def make_path(self, path):
        return os.path.join(os.path.dirname(__file__), path)

    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    # MimeType

    @pytest.mark.asyncio
    async def test_response_without_mimetype(self):
        path = self.make_path("fixtures/unknown_mime_type")
        with open(path, "rb") as f:
            body = f.read()
        app = self.make_app_with_response(path)

        async with run(app) as client:
            response = await client.get("/")
            # status
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, "OK")
            # headers
            self.assertEqual(response.headers.get("Server"), server_version)
            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "application/octet-stream")
            self.assertEqual(response.headers.get("Accept-Ranges"), "bytes")
            self.assertIsNotNone(response.headers.get("Last-Modified"))
            self.assertIsNotNone(response.headers.get("Date"))
            # body
            raw = await response.read()
            self.assertEqual(raw, body)

    @pytest.mark.asyncio
    async def test_response_with_mimetype(self):
        path = self.make_path("fixtures/users.json")
        with open(path, "rb") as f:
            body = f.read()
        app = self.make_app_with_response(path)

        async with run(app) as client:
            response = await client.get("/")
            # status
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, "OK")
            # headers
            self.assertEqual(response.headers.get("Server"), server_version)
            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "application/json")
            self.assertEqual(response.headers.get("Accept-Ranges"), "bytes")
            self.assertIsNotNone(response.headers.get("Last-Modified"))
            self.assertIsNotNone(response.headers.get("Date"))
            # body
            raw = await response.read()
            self.assertEqual(raw, body)

    # Attachment

    @pytest.mark.asyncio
    async def test_response_with_attachment(self):
        path = self.make_path("fixtures/users.json")
        app = self.make_app_with_response(path, attachment=True)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Content-Type"), "application/json")
            self.assertEqual(response.headers.get("Content-Disposition"), "attachment")

    @pytest.mark.asyncio
    async def test_response_with_attachment_and_custom_header(self):
        path = self.make_path("fixtures/users.json")
        custom_header_key, custom_header_val = "Cutom-Header", "Value"
        app = self.make_app_with_response(path, attachment=True, headers={
            custom_header_key: custom_header_val
        })

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Content-Type"), "application/json")
            self.assertEqual(response.headers.get("Content-Disposition"), "attachment")
            self.assertEqual(response.headers.get(custom_header_key), custom_header_val)

    @pytest.mark.asyncio
    async def test_response_with_attachment_and_filename(self):
        path = self.make_path("fixtures/users.json")
        app = self.make_app_with_response(path, attachment="users.json")

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Content-Type"), "application/json")
            self.assertEqual(response.headers.get("Content-Disposition"),
                             'attachment; filename="users.json"')

    # Headers

    @pytest.mark.asyncio
    async def test_response_headers(self):
        path = self.make_path("fixtures/users.json")
        custom_header_key, custom_header_val = "Cutom-Header", "Value"
        app = self.make_app_with_response(path, headers={custom_header_key: custom_header_val})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get(custom_header_key), custom_header_val)
            self.assertEqual(response.headers.get("Server"), server_version)

    @pytest.mark.asyncio
    async def test_response_with_custom_server_header(self):
        path = self.make_path("fixtures/users.json")
        server_header = "server version x"
        app = self.make_app_with_response(path, headers={"Server": server_header})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Server"), server_header)

    @pytest.mark.asyncio
    async def test_response_with_expect_header(self):
        app = self.make_app_with_response(self.make_path("fixtures/users.json"))

        async with run(app) as client:
            response = await client.post("/", json={}, expect100=True)
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_response_with_incorrect_expect_header(self):
        app = self.make_app_with_response(self.make_path("fixtures/users.json"))

        async with run(app) as client:
            response = await client.post("/", json={}, headers={"Expect": "banana"})
            self.assertEqual(response.status, 417)



import pytest
from pytest import raises

from jj.expiration_policy import ExpireAfterRequests

from .._test_utils.steps import given, then, when


async def test_not_expired_after_requests():
    with given:
        expiration_policy = ExpireAfterRequests(1)

    with when:
        is_expired = await expiration_policy.is_expired(None)

    with then:
        assert not is_expired


async def test_expired_after_requests():
    with given:
        expiration_policy = ExpireAfterRequests(1)
        await expiration_policy.is_expired(None)

    with when:
        is_expired = expiration_policy.is_expired(None)

    with then:
        assert is_expired


@pytest.mark.parametrize("count_requests", [0, -1])
def test_expired_after_requests_with_invalid_count_requests(count_requests: int):
    with when:
        with raises(Exception) as exception:
            ExpireAfterRequests(count_requests)

    with then:
        assert exception.type is AssertionError



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest
from multidict import MultiDictProxy

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher, PathMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestRequest(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    # params

    @pytest.mark.asyncio
    async def test_request_params_without_query(self):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                self.assertIsInstance(request.params, MultiDictProxy)
                self.assertEqual(request.params, {})
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_request_params_with_query(self):
        params = {"key1": "1", "key2": "2"}

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                self.assertIsInstance(request.params, MultiDictProxy)
                self.assertEqual(request.params, params)
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/", params=params)
            self.assertEqual(response.status, 200)

    # segments

    @pytest.mark.asyncio
    async def test_request_without_segments(self):
        class App(jj.App):
            resolver = self.resolver

            @PathMatcher("/users/1", resolver=resolver)
            async def handler(request):
                self.assertIsInstance(request.segments, dict)
                self.assertEqual(request.segments, {})
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/users/1")
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_request_with_segments(self):
        class App(jj.App):
            resolver = self.resolver

            @PathMatcher("/users/{user_id}", resolver=resolver)
            async def handler(request):
                self.assertIsInstance(request.segments, dict)
                self.assertEqual(request.segments, {"user_id": "1"})
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/users/1")
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_request_segments_with_other_matchers(self):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                self.assertIsInstance(request.segments, dict)
                self.assertEqual(request.segments, {})
                return Response(status=200)

        async with run(App()) as client:
            response = await client.get("/users/1")
            self.assertEqual(response.status, 200)



import asyncio
from types import TracebackType
from typing import List, Optional, Type, Union

from rtry import CancelledError, retry
from rtry.types import AttemptValue, DelayCallable, DelayValue, LoggerCallable, TimeoutValue

from ._history import HistoryItem
from ._remote_handler import RemoteHandler

__all__ = ("Mocked",)


class Mocked:
    def __init__(self, handler: RemoteHandler, *,
                 disposable: bool = True,
                 prefetch_history: bool = True) -> None:
        self._handler = handler
        self._disposable = disposable
        self._prefetch_history = prefetch_history
        self._history: Union[List[HistoryItem], None] = None

    @property
    def history(self) -> Union[List[HistoryItem], None]:
        return self._history

    async def fetch_history(self) -> List[HistoryItem]:
        self._history = await self._handler.fetch_history()
        return self._history

    async def wait_for_requests(self, count: int = 1, *,
                                timeout: TimeoutValue = 0,
                                attempts: Optional[AttemptValue] = None,
                                delay: Optional[Union[DelayValue, DelayCallable]] = None,
                                logger: Optional[LoggerCallable] = None) -> None:
        try:
            await retry(until=lambda x: len(x) < count,
                        attempts=attempts,
                        timeout=timeout,
                        delay=delay,
                        logger=logger)(self.fetch_history)()
        except CancelledError:
            pass

    async def __aenter__(self) -> "Mocked":
        self._history = None
        await self._handler.register()
        return self

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        if self._prefetch_history:
            await self.fetch_history()

        if self._disposable:
            await self._handler.deregister()

    def __enter__(self) -> "Mocked":
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return asyncio.run(self.__aenter__())
        else:
            raise RuntimeError("Use 'async with' instead")

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        return asyncio.run(self.__aexit__(exc_type, exc_val, exc_tb))

    def __repr__(self) -> str:
        return (f"Mocked<{self._handler}, "
                f"disposable={self._disposable}, "
                f"prefetch_history={self._prefetch_history}>")



from typing import Optional, cast

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ..requests import Request
from ..resolvers import Resolver
from ..responses import StreamResponse

__all__ = ("AbstractMiddleware",)


class AbstractMiddleware:
    def __init__(self, resolver: Optional[Resolver] = None) -> None:
        if resolver is None:
            resolver = getattr(self, "resolver", None)
            assert resolver is not None
        self._resolver = cast(Resolver, resolver)

    async def _do(self, request: Request, *,
                  handler: HandlerFunction, app: AbstractApp) -> StreamResponse:
        return await self.do(request, handler, app)

    async def do(self, request: Request, handler: HandlerFunction,
                 app: AbstractApp) -> StreamResponse:
        return await handler(request)



import re
from unittest.mock import sentinel

import pytest

from jj.matchers import AttributeMatcher, RegexMatcher

from ..._test_utils.steps import given, then, when


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected_factory", "actual", "res"), [
    (lambda: RegexMatcher("smth"), "smth", True),
    (lambda: RegexMatcher("smth"), "test", False),

    (lambda: RegexMatcher("[0-9]+"), "1234", True),
    (lambda: RegexMatcher("[0-9]+"), "test", False),

    # search vs match
    (lambda: RegexMatcher("smth"), "-smth-", True),
    (lambda: RegexMatcher("[0-9]+"), "-1234-", True),

    # flags
    (lambda: RegexMatcher("smth"), "SMTH", False),
    (lambda: RegexMatcher("smth", re.I), "SMTH", True),
    (lambda: RegexMatcher("[a-z]+"), "smth", True),
    (lambda: RegexMatcher("[a-z]+"), "SMTH", False),
    (lambda: RegexMatcher("[a-z]+", re.I), "SMTH", True),
])
async def test_exist_matcher(expected_factory, actual, res):
    with given:
        matcher = expected_factory()

    with when:
        actual = await matcher.match(actual)

    with then:
        assert actual is res


def test_is_instance_of_attribute_matcher():
    with given:
        matcher = RegexMatcher(".*")

    with when:
        actual = isinstance(matcher, AttributeMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize(("instance_factory", "representation"), [
    (lambda: RegexMatcher(".*"), "RegexMatcher('.*')"),

    (lambda: RegexMatcher(".*", re.I),
     f"RegexMatcher('.*', {re.I!r})"),

    (lambda: RegexMatcher(".*", re.I | re.M),
     f"RegexMatcher('.*', {re.I | re.M!r})"),
])
def test_repr(instance_factory, representation):
    with given:
        matcher = instance_factory()

    with when:
        actual = repr(matcher)

    with then:
        assert actual == representation


def test_pack():
    with given:
        pattern = ".*"
        matcher = RegexMatcher(pattern)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"pattern": pattern, "flags": 0}


def test_unpack():
    with given:
        kwargs = {
            "pattern": ".*",
            "future_field": sentinel,
        }

    with when:
        actual = RegexMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, RegexMatcher)



from packed import packable

from ._resolvable_matcher import ResolvableMatcher
from .attribute_matchers import (
    AttributeMatcher,
    ContainMatcher,
    EqualMatcher,
    ExistMatcher,
    NotContainMatcher,
    NotEqualMatcher,
    RegexMatcher,
)
from .logical_matchers import AllMatcher, AnyMatcher, LogicalMatcher
from .request_matchers import (
    HeaderMatcher,
    MethodMatcher,
    ParamMatcher,
    PathMatcher,
    RequestMatcher,
)

__all__ = (
    "AllMatcher",
    "AnyMatcher",
    "AttributeMatcher",
    "ContainMatcher",
    "contains",
    "EqualMatcher",
    "equals",
    "ExistMatcher",
    "exists",
    "HeaderMatcher",
    "LogicalMatcher",
    "MethodMatcher",
    "not_contains",
    "not_equals",
    "NotContainMatcher",
    "NotEqualMatcher",
    "ParamMatcher",
    "PathMatcher",
    "regex",
    "RegexMatcher",
    "RequestMatcher",
    "ResolvableMatcher",
)


@packable("jj.matchers.equals")
class equals(EqualMatcher):
    pass


@packable("jj.matchers.not_equals")
class not_equals(NotEqualMatcher):
    pass


@packable("jj.matchers.contains")
class contains(ContainMatcher):
    pass


@packable("jj.matchers.not_contains")
class not_contains(NotContainMatcher):
    pass


@packable("jj.matchers.regex")
class regex(RegexMatcher):
    pass


@packable("jj.matchers.exists")
class _exists(ExistMatcher):
    def __repr__(self) -> str:
        return "exists"


exists = _exists()



from ._abstract_app import AbstractApp
from ._base_app import BaseApp, create_app, define_app
from ._default_app import DefaultApp

__all__ = ("AbstractApp", "BaseApp", "DefaultApp",
           "define_app", "create_app")



import pytest

import jj
from jj.expiration_policy import ExpireAfterRequests, ExpireNever
from jj.middlewares import SelfMiddleware
from jj.mock import Mock, RemoteMock

from .._test_utils import run


@pytest.mark.asyncio
async def test_expire_never():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")
    policy = ExpireNever()

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response, expiration_policy=policy)
        await handler.register()

        for _ in range(2):
            response = await client.get("/")
            assert response.status == 200

            response_body = await response.read()
            assert response_body == b"text"

        await handler.deregister()

        response_after_deregister = await client.get("/")
        assert response_after_deregister.status == 404

        response_body_after_deregister = await response_after_deregister.read()
        assert response_body_after_deregister == b""


@pytest.mark.asyncio
@pytest.mark.parametrize("count_requests", [1, 2])
async def test_expire_after_requests(count_requests: int):
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")
    policy = ExpireAfterRequests(count_requests)

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response, expiration_policy=policy)
        await handler.register()

        for _ in range(count_requests):
            response = await client.get("/")
            assert response.status == 200

            response_body = await response.read()
            assert response_body == b"text"

        excess_response = await client.get("/")
        assert excess_response.status == 404

        excess_response_body = await excess_response.read()
        assert excess_response_body == b""


@pytest.mark.asyncio
async def test_expire_after_requests_with_request_and_deregister():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")
    policy = ExpireAfterRequests(1)

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response, expiration_policy=policy)
        await handler.register()

        response = await client.get("/")
        assert response.status == 200

        response_body = await response.read()
        assert response_body == b"text"

        await handler.deregister()

        response_after_deregister = await client.get("/")
        assert response_after_deregister.status == 404

        response_body_after_deregister = await response_after_deregister.read()
        assert response_body_after_deregister == b""


@pytest.mark.asyncio
async def test_expire_after_request_without_request_and_deregister():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")
    policy = ExpireAfterRequests(1)

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response, expiration_policy=policy)
        await handler.register()
        await handler.deregister()

        response = await client.get("/")
        assert response.status == 404

        response_body = await response.read()
        assert response_body == b""



from typing import Type

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ..logs import Logger
from ..requests import Request
from ..resolvers import Resolver
from ..responses import StreamResponse
from ._base_middleware import AppOrHandler, BaseMiddleware
from ._middleware_type import MiddlewareType

__all__ = ("LoggerMiddleware",)


class LoggerMiddleware(BaseMiddleware):
    def __init__(self, resolver: Resolver, logger: Logger) -> None:
        super().__init__(resolver)
        self._logger = logger

    def on_app(self, app: Type[AbstractApp]) -> None:
        super().on_app(app)
        self._resolver.register_attribute("logger", self._logger, app)

    def on_handler(self, handler: HandlerFunction) -> None:
        super().on_handler(handler)
        self._resolver.register_attribute("logger", self._logger, handler)

    def _register_middleware(self,
                             app_or_handler: AppOrHandler, middleware: MiddlewareType) -> None:
        old_middlewares = self._resolver.get_attribute("middlewares", app_or_handler, [])
        new_middlewares = [middleware] + old_middlewares
        self._resolver.register_attribute("middlewares", new_middlewares, app_or_handler)

    async def do(self, request: Request, handler: HandlerFunction,
                 app: AbstractApp) -> StreamResponse:
        logger = self._resolver.get_attribute("logger", handler, None)
        if logger is None:
            logger = self._resolver.get_attribute("logger", type(app), None)

        if logger:
            logger.info(request, extra={"jj_request": request})

        response = await handler(request)

        if logger:
            logger.info(response, extra={"jj_request": request, "jj_response": response})

        return response



from jj.expiration_policy import ExpireNever

from .._test_utils.steps import given, then, when


async def test_expired_never():
    with given:
        expiration_policy = ExpireNever()
        await expiration_policy.is_expired(None)

    with when:
        is_expired = expiration_policy.is_expired(None)

    with then:
        assert not is_expired



from ._abstract_middleware import AbstractMiddleware
from ._base_middleware import BaseMiddleware
from ._logger_middleware import LoggerMiddleware
from ._middleware_type import MiddlewareType
from ._root_middleware import RootMiddleware
from ._self_middleware import SelfMiddleware

__all__ = ("AbstractMiddleware", "MiddlewareType",
           "BaseMiddleware", "RootMiddleware",
           "LoggerMiddleware", "SelfMiddleware",)



from typing import Any, Dict, Union

from aiohttp.web_urldispatcher import DynamicResource
from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("RouteMatcher",)


class _Resource(DynamicResource):
    def match(self, path: str) -> Union[Dict[str, str], None]:
        return self._match(path)


@packable("jj.matchers.RouteMatcher")
class RouteMatcher(AttributeMatcher):
    def __init__(self, path: str) -> None:
        self._path = path
        self._resource = _Resource(path)

    def get_segments(self, path: str) -> Dict[str, str]:
        return self._resource.match(path) or {}

    async def match(self, path: str) -> bool:
        return self._resource.match(path) is not None

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._path!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"path": self._path}

    @classmethod
    def __unpacked__(cls, *, path: str, **kwargs: Any) -> "RouteMatcher":
        return cls(path)



from typing import Any, Callable, Optional, Tuple, Union

from packed import pack, unpack

import jj
from jj import default_app, default_handler
from jj.apps import BaseApp, create_app
from jj.expiration_policy import ExpirationPolicy
from jj.http.codes import BAD_REQUEST, OK
from jj.http.methods import ANY, DELETE, GET, POST
from jj.matchers import LogicalMatcher, RequestMatcher, ResolvableMatcher, exists
from jj.requests import Request
from jj.resolvers import Registry, Resolver
from jj.responses import RelayResponse, Response, StreamResponse

from ._history import HistoryRepository
from ._remote_response import RemoteResponseType

__all__ = ("Mock",)

MatcherType = Union[RequestMatcher, LogicalMatcher]


class _DecodeError(Exception):
    pass


class Mock(jj.App):
    def __init__(self,
                 app_factory: Callable[..., BaseApp] = create_app,
                 resolver_factory: Callable[..., Resolver] = Resolver) -> None:
        self._resolver = resolver_factory(Registry(), default_app, default_handler)
        self._app = app_factory(resolver=self._resolver)
        self._repo = HistoryRepository()

    def _decode(self, payload: bytes) -> Tuple[str, MatcherType, RemoteResponseType,
                                               Optional[ExpirationPolicy]]:
        def resolver(cls: Any, **kwargs: Any) -> Any:
            return cls.__unpacked__(**kwargs, resolver=self._resolver)

        try:
            decoded = unpack(payload, {ResolvableMatcher: resolver})
        except Exception as e:
            raise _DecodeError(f"Decode Error: can't unpack message ({e})")

        errors = []

        handler_id = decoded.get("id")
        if not isinstance(handler_id, str):
            errors.append(f"Decode Error: invalid handler id ({handler_id!r})")

        matcher = decoded.get("request")
        if not isinstance(matcher, (RequestMatcher, LogicalMatcher)):
            errors.append(f"Decode Error: invalid request field ({matcher!r})")

        response = decoded.get("response")
        if not isinstance(response, (Response, RelayResponse)):
            errors.append(f"Decode Error: invalid response field ({response!r})")

        expiration_policy = decoded.get("expiration_policy")
        if not isinstance(expiration_policy, (ExpirationPolicy, type(None))):
            errors.append(f"Decode Error: invalid expiration policy ({expiration_policy!r})")

        if len(errors) > 0:
            raise _DecodeError("\n".join(errors))

        return handler_id, matcher, response, expiration_policy

    @jj.match_any([
        jj.match(POST, "/__jj__/register", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(POST, headers={"x-jj-remote-mock": exists})
    ])
    async def register(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, matcher, response, expiration_policy = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        async def handler(request: Request) -> RemoteResponseType:
            return response.copy()

        self._resolver.register_attribute("handler_id", handler_id, handler)
        self._resolver.register_attribute("expiration_policy", expiration_policy, handler)
        setattr(self._app.__class__, handler_id, matcher(handler))

        return Response(status=OK, json={"status": OK})

    @jj.match_any([
        jj.match(DELETE, "/__jj__/deregister", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(DELETE, headers={"x-jj-remote-mock": exists})
    ])
    async def deregister(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, *_ = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        try:
            delattr(self._app.__class__, handler_id)
        except AttributeError:
            pass

        await self._repo.delete_by_tag(handler_id)

        return Response(status=OK, json={"status": OK})

    @jj.match_any([
        jj.match(GET, "/__jj__/history", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(GET, headers={"x-jj-remote-mock": exists})
    ])
    async def history(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, *_ = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        history = await self._repo.get_by_tag(handler_id)
        packed = pack(history)
        return Response(status=OK, body=packed)

    @jj.match(ANY)
    async def resolve(self, request: Request) -> StreamResponse:
        handler = await self._resolver.resolve(request, self._app)
        response = await handler(request)

        handler_id = self._resolver.get_attribute("handler_id", handler, default=None)
        if handler_id:
            await self._repo.add(request, response, tags=[handler_id])

        return response



from pathlib import Path
from typing import MutableMapping, Optional, Union

from aiohttp import web

from ._stream_response import StreamResponse

__all__ = ("StaticResponse",)


class StaticResponse(web.FileResponse, StreamResponse):
    def __init__(self, path: Union[str, Path], *,
                 chunk_size: int = 256 * 1024,
                 attachment: Union[bool, str] = False,
                 headers: Optional[MutableMapping[str, str]] = None) -> None:
        headers = headers or {}

        if isinstance(attachment, str):
            headers["Content-Disposition"] = "attachment; filename=\"{}\"".format(attachment)
        elif attachment is not False:
            headers["Content-Disposition"] = "attachment"

        super().__init__(path, chunk_size, headers=headers)



import pytest
from pytest import raises

import jj
from jj.middlewares import SelfMiddleware
from jj.mock import Mock, RemoteMock
from jj.mock._remote_mock import _RemoteMockError

from .._test_utils import run


@pytest.mark.asyncio
async def test_mock_default_handler():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)

    async with run(mock, middlewares=[self_middleware]) as client:
        response = await client.get("/")
        assert response.status == 404


@pytest.mark.asyncio
async def test_mock_register():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response)
        await handler.register()

        response = await client.get("/")
        response_body = await response.read()

        assert response.status == 200
        assert response_body == b"text"


@pytest.mark.asyncio
async def test_mock_register_bad_request():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(None, None)

        with raises(Exception) as exception:
            await handler.register()
        assert exception.type is _RemoteMockError


@pytest.mark.asyncio
async def test_mock_deregister():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response)
        await handler.register()
        await handler.deregister()

        response = await client.get("/")
        assert response.status == 404


@pytest.mark.asyncio
async def test_mock_deregister_not_registered():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(matcher, response)
        await handler.deregister()


@pytest.mark.asyncio
async def test_mock_deregister_bad_request():
    mock = Mock()
    self_middleware = SelfMiddleware(mock.resolver)

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler = remote_mock.create_handler(None, None)

        with raises(Exception) as exception:
            await handler.deregister()
        assert exception.type is _RemoteMockError


@pytest.mark.asyncio
async def test_mock_handler_order():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher1, response1 = jj.match("*"), jj.Response(body=b"text1")
    matcher2, response2 = jj.match("*"), jj.Response(body=b"text2")

    async with run(mock, middlewares=[self_middleware]) as client:
        remote_mock = RemoteMock(client.make_url("/"))
        handler1 = remote_mock.create_handler(matcher1, response1)
        handler2 = remote_mock.create_handler(matcher2, response2)
        async with handler1:
            async with handler2:
                resp = await client.get("/")
                assert resp.status == 200
                body = await resp.read()
                assert body == b"text2"



from io import SEEK_END

from aiohttp.web_request import FileField


class RequestFormatter:
    def __init__(self, request):
        self._request = request

    def format_params(self, params):
        return {key: params.getall(key) for key in params}

    def format_headers(self, headers):
        return {key: headers.getall(key) for key in headers}

    def format_data(self, data):
        return {key: data.getall(key) for key in data}

    def format_file(self, file_field):
        file_field.file.seek(0, SEEK_END)
        size = file_field.file.tell()
        file_field.file.close()
        return {
            "name": file_field.filename,
            "size": str(size),
            "type": file_field.content_type,
        }

    def format_form(self, form):
        payload = {}
        for key in form:
            payload[key] = []
            for val in form.getall(key):
                payload[key] += [self.format_file(val) if isinstance(val, FileField) else val]
        return payload

    def format_raw(self, raw):
        try:
            return raw.decode()
        except UnicodeDecodeError:
            return "<binary>"

    async def format(self):
        data = await self._request.post()
        raw = await self._request.read()
        return {
            "method": self._request.method,
            "path": self._request.path,
            "params": self.format_params(self._request.query),
            "headers": self.format_headers(self._request.headers),
            "data": self.format_data(data) if len(data) > 0 and len(raw) > 0 else None,
            "form": self.format_form(data) if len(data) > 0 and len(raw) == 0 else None,
            "raw": self.format_raw(raw),
        }



from asyncio import AbstractEventLoop
from collections import OrderedDict
from functools import partial
from typing import Any, Iterator, List

from aiohttp.http_writer import HttpVersion11
from aiohttp.web import BaseRunner
from aiohttp.web import Server as WebServer
from aiohttp.web_exceptions import HTTPExpectationFailed

from ..apps import AbstractApp
from ..middlewares import RootMiddleware
from ..requests import Request
from ..resolvers import Resolver
from ..responses import Response

__all__ = ("AppRunner",)


class AppRunner(BaseRunner):
    def __init__(self, app: AbstractApp,
                 resolver: Resolver,
                 middlewares: List[RootMiddleware],
                 loop: AbstractEventLoop,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._app = app
        self._resolver = resolver
        self._middlewares = middlewares
        self._loop = loop

    def _merge_middlewares(self, root_middlewares: List[Any],
                           app_middlewares: List[Any],
                           handler_middlewares: List[Any]) -> Iterator[Any]:
        middlewares: OrderedDict[Any, Any] = OrderedDict()

        for middleware in app_middlewares:
            middlewares[type(middleware.__self__)] = middleware

        for middleware in handler_middlewares:
            middlewares[type(middleware.__self__)] = middleware

        for middleware in root_middlewares:
            unique_key = type(middleware.__self__)
            if unique_key not in middlewares:
                middlewares[unique_key] = middleware

        return reversed(middlewares.values())

    async def _handle(self, request: Request) -> Response:
        expect = request.headers.get("EXPECT")
        if (expect is not None) and (request.version == HttpVersion11):
            if expect.lower() == "100-continue":
                await request.writer.write(b"HTTP/1.1 100 Continue\r\n\r\n")
            else:
                raise HTTPExpectationFailed()

        try:
            resolver = self._app.resolver  # type: ignore
        except (AttributeError, NotImplementedError):
            resolver = self._resolver

        handler = await resolver.resolve(request, self._app)
        unwrapped = resolver.unwrap(handler)

        root_middlewares = [middleware(handler) for middleware in self._middlewares]
        app_middlewares = resolver.get_attribute("middlewares", type(self._app), [])
        handler_middlewares = resolver.get_attribute("middlewares", unwrapped, [])

        middlewares = self._merge_middlewares(root_middlewares,
                                              app_middlewares, handler_middlewares)
        for middleware in middlewares:
            handler = partial(middleware, handler=handler, app=self._app)

        response = await handler(request)
        return response  # type: ignore

    def _make_request(self, *args: Any, **kwargs: Any) -> Request:
        return Request(*args, **kwargs, loop=self._loop)

    async def _make_server(self) -> WebServer:
        return WebServer(self._handle, request_factory=self._make_request)  # type: ignore

    async def shutdown(self) -> None:
        pass

    async def _cleanup_server(self) -> None:
        pass



import logging
from logging import LogRecord

from ...requests import Request
from ...responses import Response

__all__ = ("Formatter",)


class Formatter(logging.Formatter):
    def format_request(self, request: Request, record: LogRecord) -> str:
        return super().format(record)

    def format_response(self, response: Response, request: Request, record: LogRecord) -> str:
        return super().format(record)

    def format(self, record: LogRecord) -> str:
        request = getattr(record, "jj_request", None)
        response = getattr(record, "jj_response", None)
        if (response is not None) and (request is not None):
            return self.format_response(response, request, record)
        elif request is not None:
            return self.format_request(request, record)
        else:
            return super().format(record)



from typing import Any, Dict

from packed import packable

from ..requests import Request
from ._expiration_policy import ExpirationPolicy

__all__ = ("ExpireAfterRequests",)


@packable("jj.expiration_policy.ExpireAfterRequests")
class ExpireAfterRequests(ExpirationPolicy):
    def __init__(self, max_requests_count: int) -> None:
        assert max_requests_count > 0, \
            f'max_requests_count must be more than 0, {max_requests_count} given'

        self._max_requests_count = max_requests_count
        self._current_requests_count = 0

    async def is_expired(self, request: Request) -> bool:
        if self._current_requests_count < self._max_requests_count:
            self._current_requests_count += 1
            return False
        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._max_requests_count!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {
            "max_requests_count": self._max_requests_count,
        }

    @classmethod
    def __unpacked__(cls, *, max_requests_count: int, **kwargs: Any) -> "ExpireAfterRequests":
        return cls(max_requests_count=max_requests_count)



import io
from http.cookies import Morsel
from json import dumps
from typing import Any, Dict, List, Optional, Tuple, Union
from unittest.mock import sentinel as nil

from aiohttp import web
from aiohttp.payload import BytesPayload, IOBasePayload, TextIOPayload
from aiohttp.typedefs import LooseHeaders
from aiohttp.web import ContentCoding
from multidict import CIMultiDict
from packed import packable

from ..http.headers import CONTENT_DISPOSITION, CONTENT_TYPE
from ._stream_response import StreamResponse

__all__ = ("Response",)


@packable("jj.responses.Response")
class Response(web.Response, StreamResponse):
    def __init__(self, *,
                 json: Any = nil,
                 body: Optional[Union[str, bytes]] = None,
                 text: Optional[str] = None,
                 status: int = 200,
                 reason: Optional[str] = None,
                 headers: Optional[LooseHeaders] = None) -> None:
        headers = CIMultiDict(headers or {})

        if json is not nil:
            assert (body is None) and (text is None)
            body = dumps(json)
            if CONTENT_TYPE not in headers:
                headers[CONTENT_TYPE] = "application/json"

        if (body is None) and (text is None):
            body = ""

        if isinstance(body, io.IOBase):
            if CONTENT_DISPOSITION not in headers:
                headers[CONTENT_DISPOSITION] = "inline"

        super().__init__(body=body, text=text, status=status, reason=reason, headers=headers)

    @property
    def content_coding(self) -> Optional[ContentCoding]:
        return self._compression_force

    def _cookie_to_dict(self, cookie: "Morsel[str]") -> Dict[str, Union[str, None]]:
        dictionary: Dict[str, Union[str, None]] = {
            "name": cookie.key,
            "value": cookie.value,
        }
        for attr in ("expires", "domain", "max-age", "path", "secure", "httponly", "version"):
            key = attr.replace("-", "_")
            val = cookie.get(attr)
            dictionary[key] = None if val == "" else val
        return dictionary

    def copy(self) -> "Response":
        assert not self.prepared

        response = self.__class__(status=self.status, reason=self.reason,
                                  headers=self.headers, body=self.body)  # type: ignore
        for cookie in self.cookies.values():
            response.set_cookie(**self._cookie_to_dict(cookie))  # type: ignore
        if self.chunked:
            response.enable_chunked_encoding()
        if self._compression_force:
            response.enable_compression(self._compression_force)
        return response

    def get_body(self) -> bytes:
        if isinstance(self.body, (bytes, bytearray, memoryview)):
            return bytes(self.body)
        elif isinstance(self.body, BytesPayload):
            return bytes(self.body._value)
        elif isinstance(self.body, TextIOPayload):
            return bytes(self.body._value.read(), self.body.encoding)  # type: ignore
        elif isinstance(self.body, IOBasePayload):
            return bytes(self.body._value.read())
        else:
            raise ValueError("Unsupported body type {}".format(type(self.body)))

    def __packed__(self) -> Dict[str, Any]:
        assert not self.prepared

        headers = [[key, val] for key, val in self.headers.items()]
        cookies = [self._cookie_to_dict(cookie) for cookie in self.cookies.values()]

        body = self.get_body()

        compression: Optional[str] = None
        if isinstance(self._compression_force, ContentCoding):
            compression = self._compression_force.value
        else:
            compression = self._compression_force

        return {
            "status": self.status,
            "reason": self.reason,
            "headers": headers,
            "cookies": cookies,
            "body": body,
            "chunked": self.chunked,
            "compression": compression,
        }

    @classmethod
    def __unpacked__(cls, *,
                     status: int,
                     reason: Optional[str],
                     headers: List[Tuple[str, str]],
                     cookies: List[Dict[str, Union[str, None]]],
                     body: Optional[bytes],
                     chunked: bool,
                     compression: Optional[ContentCoding],
                     **kwargs: Any) -> "Response":
        response = cls(status=status, reason=reason, headers=headers, body=body)  # type: ignore
        for cookie in cookies:
            response.set_cookie(**cookie)  # type: ignore
        if compression:
            response.enable_compression(compression)
        if chunked:
            response.enable_chunked_encoding()
        return response



import re
from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("RegexMatcher",)


@packable("jj.matchers.RegexMatcher")
class RegexMatcher(AttributeMatcher):
    def __init__(self, pattern: str, flags: int = 0) -> None:
        self._pattern = pattern
        self._flags = flags
        self._compiled = re.compile(self._pattern, self._flags)

    async def match(self, actual: Any) -> bool:
        return self._compiled.search(actual) is not None

    def __repr__(self) -> str:
        if self._flags == 0:
            return f"{self.__class__.__qualname__}({self._pattern!r})"
        return f"{self.__class__.__qualname__}({self._pattern!r}, {self._flags!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"pattern": self._pattern, "flags": self._flags}

    @classmethod
    def __unpacked__(cls, *, pattern: str, flags: int = 0, **kwargs: Any) -> "RegexMatcher":
        return cls(pattern, flags)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest
from aiohttp import FormData

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.http.methods import DELETE, GET, POST
from jj.matchers import MethodMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import RelayResponse, Response

from .._test_utils import TestServer, run
from ._request_formatter import RequestFormatter


class TestRelayResponse(TestCase):
    def make_app_with_response(self, *args, **kwargs):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return RelayResponse(*args, **kwargs)
        return App()

    def make_debug_app(self):
        @MethodMatcher("*", resolver=self.resolver)
        async def handler(request):
            payload = await RequestFormatter(request).format()
            return Response(json=payload)
        return create_app(resolver=self.resolver, handlers={"handler": handler})

    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    # Method

    @pytest.mark.asyncio
    async def test_get_request(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)

            async with run(app) as client:
                response = await client.request(GET, "/")
                body = await response.json()
                self.assertEqual(body["method"], GET)

    @pytest.mark.asyncio
    async def test_delete_request(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)

            async with run(app) as client:
                response = await client.request(DELETE, "/")
                body = await response.json()
                self.assertEqual(body["method"], DELETE)

    # Body

    @pytest.mark.asyncio
    async def test_post_request_with_no_data(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)

            async with run(app) as client:
                response = await client.request(POST, "/")
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["data"], None)

    @pytest.mark.asyncio
    async def test_request_with_post_data(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            payload = FormData([
                ("field1", "value1"),
                ("field1", "value2"),
                ("field2", "null"),
            ])

            async with run(app) as client:
                response = await client.request(POST, "/", data=payload)
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["data"], {
                    "field1": ["value1", "value2"],
                    "field2": ["null"],
                })

    @pytest.mark.asyncio
    async def test_request_with_form_data(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            payload = FormData([
                ("field1", "value1"),
                ("field1", "value2"),
                ("field2", "null"),
            ])
            payload._is_multipart = True

            async with run(app) as client:
                response = await client.request(POST, "/", data=payload)
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["form"], {
                    "field1": ["value1", "value2"],
                    "field2": ["null"],
                })

    @pytest.mark.asyncio
    async def test_request_with_file(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            payload = FormData()
            payload.add_field("field1", "value1")
            payload.add_field("field2", b"binary", content_type="image/jpeg", filename="image.jpg")

            async with run(app) as client:
                response = await client.request(POST, "/", data=payload)
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["form"], {
                    "field1": ["value1"],
                    "field2": [{
                        "name": "image.jpg",
                        "size": str(len(b"binary")),
                        "type": "image/jpeg",
                    }],
                })

    @pytest.mark.asyncio
    async def test_request_with_json_data(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            payload = {"field1": "value1", "field2": None}
            payload_serialized = '{"field1": "value1", "field2": null}'

            async with run(app) as client:
                response = await client.request(POST, "/", json=payload)
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["headers"].get("Content-Type"), ["application/json"])
                self.assertEqual(body["raw"], payload_serialized)

    @pytest.mark.asyncio
    async def test_request_with_binary_data(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)

            async with run(app) as client:
                response = await client.request(POST, "/", data=b"binary")
                body = await response.json()
                self.assertEqual(body["method"], POST)
                self.assertEqual(body["headers"].get("Content-Type"), ["application/octet-stream"])
                self.assertEqual(body["raw"], "binary")

    # Path

    @pytest.mark.asyncio
    async def test_request_with_custom_path(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            path = "/users/1234"

            async with run(app) as client:
                response = await client.request(GET, path)
                body = await response.json()
                self.assertEqual(body["method"], GET)
                self.assertEqual(body["path"], path)

    # Params

    @pytest.mark.asyncio
    async def test_request_with_custom_query_params(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            params = [
                ("field1", "value1"),
                ("field1", "value2"),
                ("field2", "null"),
            ]

            async with run(app) as client:
                response = await client.request(GET, "/", params=params)
                body = await response.json()
                self.assertEqual(body["method"], GET)
                self.assertEqual(body["params"], {
                    "field1": ["value1", "value2"],
                    "field2": ["null"],
                })

    # Headers

    @pytest.mark.asyncio
    async def test_request_with_custom_headers(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)
            headers = [
                ("x-header-1", "value1"),
                ("x-header-1", "value2"),
                ("x-header-2", "null"),
            ]

            async with run(app) as client:
                response = await client.request(GET, "/", headers=headers)
                body = await response.json()
                self.assertEqual(body["method"], GET)
                # aiohttp client не умеет отправлять несколько заголовков с одним названием,
                # однако aiohttp server умеет их принимать
                self.assertEqual(body["headers"].get("x-header-1"), ["value2"])
                self.assertEqual(body["headers"].get("x-header-2"), ["null"])

    @pytest.mark.asyncio
    async def test_request_default_headers(self):
        debug_app = self.make_debug_app()
        async with TestServer(debug_app) as server:
            url = str(server.make_url("/"))
            app = self.make_app_with_response(target=url)

            async with run(app) as client:
                response = await client.request(GET, "/")
                body = await response.json()
                self.assertEqual(body["headers"].get("Host"),
                                 ["{}:{}".format(server.host, server.port)])
                self.assertEqual(body["headers"].get("Accept"), ["*/*"])
                self.assertEqual(body["headers"].get("Accept-Encoding"), ["gzip, deflate"])
                self.assertEqual(body["headers"].get("Content-Type"), ["application/octet-stream"])
                self.assertEqual(body["headers"].get("Transfer-Encoding"), ["chunked"])
                self.assertIsNotNone(body["headers"].get("User-Agent"))
                self.assertEqual(len(body["headers"]), 6)



# flake8: noqa
from pytest import raises

# apps


def test_import_abstract_app():
    from jj.apps import AbstractApp
    with raises(ImportError):
        from jj import AbstractApp


def test_import_base_app():
    from jj.apps import BaseApp
    with raises(ImportError):
        from jj import BaseApp


def test_import_default_app():
    from jj.apps import DefaultApp
    with raises(ImportError):
        from jj import DefaultApp


def test_import_app_factory():
    from jj.apps import define_app, create_app
    with raises(ImportError):
        from jj import define_app, create_app


def test_import_registrar():
    with raises(ImportError):
        from jj.apps import _Registrar


def test_import_singleton():
    with raises(ImportError):
        from jj.apps import _Singleton


def test_import_app():
    from jj import App


# handlers


def test_import_handler_function():
    from jj.handlers import HandlerFunction
    with raises(ImportError):
        from jj import HandlerFunction


def test_import_default_handler():
    from jj.handlers import default_handler
    from jj import default_handler


# http


def test_import_http_codes():
    from jj.http.codes import OK
    from jj.http import OK
    with raises(ImportError):
        from jj import OK


def test_import_http_headers():
    from jj.http.headers import CONTENT_TYPE
    from jj.http import CONTENT_TYPE
    with raises(ImportError):
        from jj import CONTENT_TYPE


def test_import_http_methods():
    from jj.http.methods import GET
    from jj.http import GET
    with raises(ImportError):
        from jj import GET


# logs


def test_import_logger():
    from jj.logs import Logger
    with raises(ImportError):
        from jj import Logger


def test_import_default_logger():
    from jj.logs import logger
    with raises(ImportError):
        from jj import logger
    from jj import default_logger


def test_import_filter():
    from jj.logs import Filter
    with raises(ImportError):
        from jj import Filter


def test_import_default_filter():
    from jj.logs import filter_
    with raises(ImportError):
        from jj import filter_


def test_import_formatter():
    from jj.logs import Formatter
    with raises(ImportError):
        from jj import Formatter


def test_import_default_formatter():
    from jj.logs import formatter
    with raises(ImportError):
        from jj import formatter


def test_import_simple_formatter():
    from jj.logs import SimpleFormatter
    with raises(ImportError):
        from jj import SimpleFormatter


def test_import_default_log_handler():
    from jj.logs import handler
    with raises(ImportError):
        from jj import handler


# matchers


def test_import_resolvable_matcher():
    from jj.matchers import ResolvableMatcher
    with raises(ImportError):
        from jj import ResolvableMatcher


def test_import_match_methods():
    from jj import match_method
    from jj import match_path
    from jj import match_header, match_headers
    from jj import match_param, match_params
    from jj import match_all, match_any
    from jj import match


def test_import_matcher_shorthands():
    from jj.matchers import exists
    with raises(ImportError):
        from jj import exists

    from jj.matchers import equals, not_equals
    with raises(ImportError):
        from jj import equals
    with raises(ImportError):
        from jj import not_equals

    from jj.matchers import contains, not_contains
    with raises(ImportError):
        from jj import contains
    with raises(ImportError):
        from jj import not_contains

    from jj.matchers import regex
    with raises(ImportError):
        from jj import regex


# matchers/attribute_matchers


def test_import_attribute_matcher():
    from jj.matchers.attribute_matchers import AttributeMatcher
    from jj.matchers import AttributeMatcher
    with raises(ImportError):
        from jj import AttributeMatcher


def test_import_attribute_contain_matcher():
    from jj.matchers.attribute_matchers import ContainMatcher
    from jj.matchers import ContainMatcher
    with raises(ImportError):
        from jj import ContainMatcher


def test_import_attribute_not_contain_matcher():
    from jj.matchers.attribute_matchers import NotContainMatcher
    from jj.matchers import NotContainMatcher
    with raises(ImportError):
        from jj import NotContainMatcher


def test_import_attribute_equal_matcher():
    from jj.matchers.attribute_matchers import EqualMatcher
    from jj.matchers import EqualMatcher
    with raises(ImportError):
        from jj import EqualMatcher


def test_import_attribute_not_equal_matcher():
    from jj.matchers.attribute_matchers import NotEqualMatcher
    from jj.matchers import NotEqualMatcher
    with raises(ImportError):
        from jj import NotEqualMatcher


def test_import_attribute_exist_matcher():
    from jj.matchers.attribute_matchers import ExistMatcher
    from jj.matchers import ExistMatcher
    with raises(ImportError):
        from jj import ExistMatcher


def test_import_attribute_multi_dict_matcher():
    from jj.matchers.attribute_matchers import MultiDictMatcher
    with raises(ImportError):
        from jj.matchers import MultiDictMatcher
    with raises(ImportError):
        from jj import MultiDictMatcher


def test_import_attribute_regex_matcher():
    from jj.matchers.attribute_matchers import RegexMatcher
    from jj.matchers import RegexMatcher
    with raises(ImportError):
        from jj import RegexMatcher


def test_import_attribute_route_matcher():
    from jj.matchers.attribute_matchers import RouteMatcher
    with raises(ImportError):
        from jj.matchers import RouteMatcher
    with raises(ImportError):
        from jj import RouteMatcher


# matchers/logical_matchers


def test_import_logical_all_matcher():
    from jj.matchers.logical_matchers import AllMatcher
    from jj.matchers import AllMatcher
    with raises(ImportError):
        from jj import AllMatcher


def test_import_logical_any_matcher():
    from jj.matchers.logical_matchers import AnyMatcher
    from jj.matchers import AnyMatcher
    with raises(ImportError):
        from jj import AnyMatcher


def test_import_logical_matcher():
    from jj.matchers.logical_matchers import LogicalMatcher
    from jj.matchers import LogicalMatcher
    with raises(ImportError):
        from jj import LogicalMatcher


# matchers/request_matchers


def test_import_request_header_matcher():
    from jj.matchers.request_matchers import HeaderMatcher
    from jj.matchers import HeaderMatcher
    with raises(ImportError):
        from jj import HeaderMatcher


def test_import_request_method_matcher():
    from jj.matchers.request_matchers import MethodMatcher
    from jj.matchers import MethodMatcher
    with raises(ImportError):
        from jj import MethodMatcher


def test_import_request_param_matcher():
    from jj.matchers.request_matchers import ParamMatcher
    from jj.matchers import ParamMatcher
    with raises(ImportError):
        from jj import ParamMatcher


def test_import_request_path_matcher():
    from jj.matchers.request_matchers import PathMatcher
    from jj.matchers import PathMatcher
    with raises(ImportError):
        from jj import PathMatcher


def test_import_request_matcher():
    from jj.matchers.request_matchers import RequestMatcher
    from jj.matchers import RequestMatcher
    with raises(ImportError):
        from jj import RequestMatcher


# middlewares


def test_import_abstract_middleware():
    from jj.middlewares import AbstractMiddleware
    with raises(ImportError):
        from jj import AbstractMiddleware


def test_import_root_middleware():
    from jj.middlewares import RootMiddleware
    with raises(ImportError):
        from jj import RootMiddleware


def test_import_base_middleware():
    from jj.middlewares import BaseMiddleware
    with raises(ImportError):
        from jj import BaseMiddleware


def test_import_self_middleware():
    from jj.middlewares import SelfMiddleware
    with raises(ImportError):
        from jj import SelfMiddleware


def test_import_logger_middleware():
    from jj.middlewares import LoggerMiddleware
    with raises(ImportError):
        from jj import LoggerMiddleware


def test_import_middleware():
    from jj import Middleware


# requests


def test_import_request():
    from jj.requests import Request
    from jj import Request


# resolvers


def test_import_registry():
    from jj.resolvers import Registry
    with raises(ImportError):
        from jj import Registry


def test_import_resolver():
    from jj.resolvers import Resolver
    with raises(ImportError):
        from jj import Resolver


def test_import_reversed_resolver():
    from jj.resolvers import ReversedResolver
    with raises(ImportError):
        from jj import ReversedResolver


# responses


def test_import_response():
    from jj.responses import Response
    from jj import Response


def test_import_static_response():
    from jj.responses import StaticResponse
    from jj import StaticResponse


def test_import_stream_response():
    from jj.responses import StreamResponse
    from jj import StreamResponse


def test_import_relay_response():
    from jj.responses import RelayResponse
    from jj import RelayResponse


# runners


def test_import_runner():
    from jj.runners import AppRunner
    with raises(ImportError):
        from jj import AppRunner


# servers


def test_import_server():
    from jj.servers import Server
    with raises(ImportError):
        from jj import Server


def test_import_start():
    from jj import start


def test_import_wait_for():
    from jj import wait_for


def test_import_serve():
    from jj import serve


def test_import_version():
    from jj import version
    from jj import server_version



from typing import Any, Dict

from packed import packable

from ..requests import Request
from ._expiration_policy import ExpirationPolicy

__all__ = ("ExpireNever",)


@packable("jj.expiration_policy.ExpireNever")
class ExpireNever(ExpirationPolicy):
    async def is_expired(self, request: Request) -> bool:
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def __packed__(self) -> Dict[str, Any]:
        return dict()

    @classmethod
    def __unpacked__(cls, **kwargs: Any) -> "ExpireNever":
        return cls()



from typing import Any, Dict

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, RouteMatcher, StrOrAttrMatcher
from ._request_matcher import RequestMatcher

__all__ = ("PathMatcher",)


@packable("jj.matchers.PathMatcher")
class PathMatcher(RequestMatcher):
    def __init__(self, path: StrOrAttrMatcher, *, resolver: Resolver) -> None:
        super().__init__(resolver=resolver)
        if isinstance(path, AttributeMatcher):
            self._matcher = path
        else:
            self._matcher = RouteMatcher(path)

    async def match(self, request: Request) -> bool:
        matched = await self._matcher.match(request.path)
        if matched and isinstance(self._matcher, RouteMatcher):
            request.segments = self._matcher.get_segments(request.path)
        else:
            request.segments = None  # type: ignore
        return matched

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"path": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     path: StrOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "PathMatcher":
        return cls(path, resolver=resolver)



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest
from multidict import MultiDict

from jj.matchers import AttributeMatcher, ParamMatcher, RequestMatcher
from jj.matchers.attribute_matchers import MultiDictMatcher

from ..._test_utils.fixtures import request_, resolver_
from ..._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_",)


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ({}, {}, True),
    ([], {}, True),
    ({}, {"key": "1"}, True),
    ([], {"key": "1"}, True),

    ({"key": "1"}, {}, False),
    ({"key": "1"}, {"key": "1"}, True),
    ({"key1": "1"}, {"key1": "1", "key2": "2"}, True),
    ({"key1": "1", "key2": "2"}, {"key1": "1"}, False),
    ({"key1": "1", "key2": "2"}, {"key1": "1", "key2": "2"}, True),

    ([("key", "1"), ("key", "2")], [], False),
    ([("key", "1"), ("key", "2")], [("key", "1")], False),
    ([("key", "1"), ("key", "2")], [("key", "2")], False),
    ([("key", "1"), ("key", "2")], [("key", "1"), ("key", "2")], True),
    ([("key", "1"), ("key", "2")], [("key", "1"), ("key", "3")], False),

    ({"key": "1"}, [("key", "1")], True),
    ({"key": "1"}, [("key", "1"), ("key", "2")], True),
    ({"key": "1"}, [("key", "3"), ("key", "2")], False),

    ({"a": "b"}, {"a": "b"}, True),
    ({"a": "b"}, {"A": "b"}, False),
    ({"A": "b"}, {"a": "b"}, False),
    ({"a": "b"}, {"a": "B"}, False),
    ({"a": "B"}, {"a": "b"}, False),
])
async def test_param_matcher(expected, actual, res, *, resolver_, request_):
    with given:
        request_.query = MultiDict(actual)
        matcher = ParamMatcher(expected, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_val", "params"), [
    (True, {}),
    (False, {"key": "val"}),
    (True, [("key", "1"), ("key", "2")]),
    (False, [("key1", "1"), ("key2", "2")]),
])
async def test_param_matcher_with_custom_submatcher(ret_val, params, *, resolver_, request_):
    with given:
        request_.query = MultiDict(params)
        submatcher_ = Mock(AttributeMatcher, match=AsyncMock(return_value=ret_val))
        matcher = ParamMatcher(submatcher_, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is ret_val
        assert submatcher_.mock_calls == [call.match(request_.query)]


@pytest.mark.asyncio
async def test_param_matcher_with_value_submatchers_superset(request_):
    with given:
        request_.query = MultiDict([
            ("key1", "1.1"),
            ("key1", "1.2"),
        ])
        submatcher1_ = Mock(AttributeMatcher, match=AsyncMock(return_value=True))
        submatcher2_ = Mock(AttributeMatcher)
        matcher = ParamMatcher(resolver=resolver_, params={
            "key1": submatcher1_,
            "key2": submatcher2_,
        })

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is False
        assert submatcher1_.mock_calls == [call.match("1.1")]
        assert submatcher2_.mock_calls == []


@pytest.mark.asyncio
async def test_param_matcher_with_value_submatchers_subset(request_):
    with given:
        request_.query = MultiDict([
            ("key1", "1"),
            ("key2", "2.1"),
            ("key2", "2.2"),
            ("key3", "3"),
        ])
        submatcher1_ = Mock(AttributeMatcher, match=AsyncMock(side_effect=(True,)))
        submatcher2_ = Mock(AttributeMatcher, match=AsyncMock(side_effect=(False, True)))
        matcher = ParamMatcher(resolver=resolver_, params={
            "key1": submatcher1_,
            "key2": submatcher2_,
        })

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is True
        assert submatcher1_.mock_calls == [call.match("1")]
        assert submatcher2_.mock_calls == [call.match("2.1"), call.match("2.2")]


def test_is_instance_of_request_matcher(*, resolver_):
    with given:
        matcher = ParamMatcher({}, resolver=resolver_)

    with when:
        actual = isinstance(matcher, RequestMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize(("params", "representation"), [
    ({}, "[]"),
    ({"KEY": "Val"}, "[('KEY', 'Val')]"),
    ({"key1": "1", "key2": "2"}, "[('key1', '1'), ('key2', '2')]"),
])
def test_repr(params, representation, *, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = ParamMatcher(params, resolver=resolver_)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == f"ParamMatcher(MultiDictMatcher({representation}), resolver=<Resolver>)"


def test_pack(*, resolver_):
    with given:
        submatcher = MultiDictMatcher([("key", "1"), ("key", "2")])
        matcher = ParamMatcher(submatcher, resolver=resolver_)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"params": submatcher}


def test_unpack(*, resolver_):
    with given:
        submatcher = MultiDictMatcher([("key", "1"), ("key", "2")])
        kwargs = {
            "params": submatcher,
            "resolver": resolver_,
            "future_field": sentinel,
        }

    with when:
        actual = ParamMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, ParamMatcher)



from ..handlers import HandlerFunction
from ..requests import Request
from ..resolvers import Resolver

__all__ = ("ResolvableMatcher",)


class ResolvableMatcher:
    def __init__(self, *, resolver: Resolver) -> None:
        self._resolver = resolver

    async def match(self, request: Request) -> bool:
        raise NotImplementedError()

    def __call__(self, handler: HandlerFunction) -> HandlerFunction:
        self._resolver.register_matcher(self.match, handler)
        return handler

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(resolver={self._resolver!r})"



from typing import Optional

from aiohttp import web
from aiohttp.typedefs import LooseHeaders
from multidict import CIMultiDict

from .._version import server_version

__all__ = ("StreamResponse",)


class StreamResponse(web.StreamResponse):
    def __init__(self, *,
                 status: int = 200,
                 reason: Optional[str] = None,
                 headers: Optional[LooseHeaders] = None) -> None:
        headers = CIMultiDict(headers or {})
        headers["Server"] = headers.get("Server", server_version)
        super().__init__(status=status, reason=reason, headers=headers)



import pytest

import jj
from jj.expiration_policy import ExpireAfterRequests
from jj.middlewares import SelfMiddleware
from jj.mock import Mock, RemoteMock

from .._test_utils import run


@pytest.mark.asyncio
async def test_expire_after_requests_history_request():
    mock = Mock()
    self_middleware = SelfMiddleware(Mock().resolver)
    matcher, response = jj.match("*"), jj.Response(status=200, body=b"text")

    requests_count = 2
    policy = ExpireAfterRequests(requests_count)

    async with run(mock, middlewares=[self_middleware]) as client:
        handler = RemoteMock(client.make_url("/")).create_handler(
            matcher, response, expiration_policy=policy
        )
        async with handler:
            for _ in range(requests_count):
                resp = await client.get("/", params={"key": "val"})
                assert resp.status == 200

            excess_response = await client.get("/")
            assert excess_response.status == 404

            history = await handler.fetch_history()
            assert len(history) == requests_count

            for i in range(requests_count):
                req = history[i]["request"]
                assert req.method == "GET"
                assert req.segments == {}
                assert req.path == "/"
                assert req.params == {"key": "val"}
                assert req.headers.get("User-Agent")
                assert req.body == req.raw == b""

                res = history[i]["response"]
                assert res.status == 200
                assert res.headers.get("Server")
                assert res.reason == "OK"
                assert res.body == res.raw == b"text"



from unittest.mock import Mock, call, sentinel

import pytest

from jj.logs import Formatter

from .._test_utils.steps import given, then, when
from ._log_record import TestLogRecord


class _Formatter(Formatter):
    format_request: Mock
    format_response: Mock


@pytest.fixture()
def formatter():
    formatter = Formatter()
    formatter.format_request = Mock(return_value=sentinel.formatted_request)
    formatter.format_response = Mock(return_value=sentinel.formatted_response)
    return formatter


@pytest.fixture()
def record():
    return TestLogRecord(sentinel.message)


def test_format_without_request_and_response(formatter: _Formatter, record: TestLogRecord):
    with when:
        res = formatter.format(record)

    with then:
        assert res == str(record.msg)
        assert formatter.format_request.mock_calls == []
        assert formatter.format_response.mock_calls == []


def test_format_with_request(formatter: _Formatter, record: TestLogRecord):
    with given:
        record.jj_request = sentinel.request

    with when:
        res = formatter.format(record)

    with then:
        assert res == formatter.format_request.return_value
        assert formatter.format_request.mock_calls == [
            call(sentinel.request, record)
        ]
        assert formatter.format_response.mock_calls == []


def test_format_with_response(formatter: _Formatter, record: TestLogRecord):
    with given:
        record.jj_request = sentinel.request
        record.jj_response = sentinel.response

    with when:
        res = formatter.format(record)

    with then:
        assert res == formatter.format_response.return_value
        assert formatter.format_request.mock_calls == []
        assert formatter.format_response.mock_calls == [
            call(sentinel.response, sentinel.request, record)
        ]


def test_request_format(record: TestLogRecord):
    with given:
        formatter = Formatter()

    with when:
        res = formatter.format_request(sentinel.request, record)

    with then:
        assert res == str(record.msg)


def test_response_format(record: TestLogRecord):
    with given:
        formatter = Formatter()

    with when:
        res = formatter.format_response(sentinel.response, sentinel.request, record)

    with then:
        assert res == str(record.msg)



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest
from multidict import MultiDict

from jj.matchers import AttributeMatcher
from jj.matchers.attribute_matchers import MultiDictMatcher

from ..._test_utils.steps import given, then, when


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ({}, {}, True),
    ({}, {"key1": "1"}, True),
    ({"key1": "1"}, {}, False),
    ({"key1": "1"}, {"key1": "1"}, True),

    ({"key1": "1"}, {"key1": "1", "key2": "2"}, True),
    ({"key1": "1", "key2": "2"}, {"key1": "1"}, False),

    ({"key1": "1.1"}, [("key1", "1.1"), ("key1", "1.2")], True),
    ({"key1": "1.2"}, [("key1", "1.1"), ("key1", "1.2")], True),
    ([("key1", "1.1"), ("key1", "1.2")], [("key1", "1.1"), ("key1", "1.2")], True),
])
async def test_multi_dict_matcher(expected, actual, res):
    with given:
        matcher = MultiDictMatcher(expected)

    with when:
        actual = await matcher.match(MultiDict(actual))

    with then:
        assert actual is res


@pytest.mark.asyncio
async def test_multi_dict_matcher_with_value_submatchers_superset():
    with given:
        submatcher1_ = Mock(AttributeMatcher, match=AsyncMock(return_value=True))
        submatcher2_ = Mock(AttributeMatcher)
        matcher = MultiDictMatcher({
            "key1": submatcher1_,
            "key2": submatcher2_,
        })

    with when:
        actual = await matcher.match(MultiDict([
            ("key1", "1.1"),
            ("key1", "1.2"),
        ]))

    with then:
        assert actual is False
        assert submatcher1_.mock_calls == [call.match("1.1")]
        assert submatcher2_.mock_calls == []


@pytest.mark.asyncio
async def test_multi_dict_matcher_with_value_submatchers_subset():
    with given:
        submatcher1_ = Mock(AttributeMatcher, match=AsyncMock(side_effect=(True,)))
        submatcher2_ = Mock(AttributeMatcher, match=AsyncMock(side_effect=(False, True)))
        matcher = MultiDictMatcher({
            "key1": submatcher1_,
            "key2": submatcher2_,
        })

    with when:
        actual = await matcher.match(MultiDict([
            ("key1", "1"),
            ("key2", "2.1"),
            ("key2", "2.2"),
            ("key3", "3"),
        ]))

    with then:
        assert actual is True
        assert submatcher1_.mock_calls == [call.match("1")]
        assert submatcher2_.mock_calls == [call.match("2.1"), call.match("2.2")]


def test_is_instance_of_attribute_matcher():
    with given:
        matcher = MultiDictMatcher({})

    with when:
        actual = isinstance(matcher, AttributeMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize(("expected", "representation"), [
    ({}, "MultiDictMatcher([])"),
    ({"key": "val"}, "MultiDictMatcher([('key', 'val')])"),
    ({"key1": "1", "key2": "2"}, "MultiDictMatcher([('key1', '1'), ('key2', '2')])"),
])
def test_repr(expected, representation):
    with given:
        matcher = MultiDictMatcher(expected)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == representation


def test_pack():
    with given:
        expected = [["key", "1"], ["key", "2"]]
        matcher = MultiDictMatcher(expected)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"expected": expected}


def test_unpack():
    with given:
        kwargs = {
            "expected": [("key", "1"), ("key", "2")],
            "future_field": sentinel,
        }

    with when:
        actual = MultiDictMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, MultiDictMatcher)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

from unittest.mock import Mock, call, sentinel

import pytest

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher, ParamMatcher, PathMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestApp(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    @pytest.mark.asyncio
    async def test_app_without_handlers(self):
        class App(jj.App):
            resolver = self.resolver

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_app_with_default_handler(self):
        status, text = 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        async with run(App()) as client:
            response = await client.get("/")
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

    @pytest.mark.asyncio
    async def test_app_with_single_handler(self):
        path, status, text = "/route", 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        async with run(App()) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

            response2 = await client.get("/")
            self.assertEqual(response2.status, 404)

    @pytest.mark.asyncio
    async def test_app_with_multiple_handlers(self):
        path1, status1, text1 = "/route", 201, "text-1"
        path2, status2, text2 = "/route/subroute", 202, "text-2"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path1, resolver=resolver)
            async def handler1(request):
                return Response(status=status1, text=text1)

            @PathMatcher(path2, resolver=resolver)
            async def handler2(request):
                return Response(status=status2, text=text2)

        async with run(App()) as client:
            response1 = await client.get(path1)
            self.assertEqual(response1.status, status1)
            self.assertEqual(await response1.text(), text1)

            response2 = await client.get(path2)
            self.assertEqual(response2.status, status2)
            self.assertEqual(await response2.text(), text2)

    @pytest.mark.asyncio
    async def test_app_handlers_priority(self):
        path = "/route"
        status1, text1 = 201, "text-1"
        status2, text2 = 202, "text-2"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler1(request):
                return Response(status=status1, text=text1)

            @PathMatcher(path, resolver=resolver)
            async def handler2(request):
                return Response(status=status2, text=text2)

        async with run(App()) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status1)
            self.assertEqual(await response.text(), text1)

    @pytest.mark.asyncio
    async def test_handler_with_matcher_and_decorators(self):
        mock = Mock()

        def decorator_before(fn):
            async def before_handler(request):
                mock(sentinel.BEFORE)
                return await fn(request)
            return before_handler

        def decorator_after(fn):
            async def after_handler(request):
                mock(sentinel.AFTER)
                return await fn(request)
            return after_handler

        path, status, text = "/route", 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @decorator_after
            @PathMatcher(path, resolver=resolver)
            @decorator_before
            async def handler(request):
                mock(sentinel.HANDLE)
                return Response(status=status, text=text)

        async with run(App()) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

            response2 = await client.get("/")
            self.assertEqual(response2.status, 404)

        mock.assert_has_calls([
            call(sentinel.AFTER),
            call(sentinel.BEFORE),
            call(sentinel.HANDLE)
        ])
        self.assertEqual(mock.call_count, 3)

    @pytest.mark.asyncio
    async def test_handler_with_matchers_and_decorator(self):
        mock = Mock()

        def decorator(fn):
            async def handler(request):
                mock()
                return await fn(request)
            return handler

        status, text = 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @ParamMatcher({"key2": "2"}, resolver=resolver)
            @decorator
            @ParamMatcher({"key1": "1"}, resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        async with run(App()) as client:
            response0 = await client.get("/", params={})
            self.assertEqual(response0.status, 404)

            response1 = await client.get("/", params={"key1": "1"})
            self.assertEqual(response1.status, 404)

            response2 = await client.get("/", params={"key1": "1", "key2": "2"})
            self.assertEqual(response2.status, status)
            self.assertEqual(await response2.text(), text)

    @pytest.mark.asyncio
    async def test_app_inheritance(self):
        path, status, text = "/route", 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        class AnotherApp(App):
            pass

        async with run(AnotherApp()) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

            response2 = await client.get("/")
            self.assertEqual(response2.status, 404)

    @pytest.mark.asyncio
    async def test_app_inheritance_with_handler_overriding(self):
        path = "/route"
        status1, text1 = 201, "text-1"
        status2, text2 = 202, "text-2"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=self.resolver)
            async def handler(request):
                return Response(status=status1, text=text1)

        class AnotherApp(App):
            @PathMatcher(path, resolver=self.resolver)
            async def handler(request):
                return Response(status=status2, text=text2)

        async with run(AnotherApp(), self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status2)
            self.assertEqual(await response.text(), text2)

        async with run(App(), self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status1)
            self.assertEqual(await response.text(), text1)

    @pytest.mark.asyncio
    async def test_app_inheritance_with_default_handler_overriding(self):
        status1, text1 = 201, "text-1"
        status2, text2 = 202, "text-2"
        status3, text3 = 203, "text-3"

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=self.resolver)
            async def other(request):
                return Response(status=status1, text=text1)

        class AnotherApp(App):
            @PathMatcher("/route", resolver=self.resolver)
            async def handler(request):
                return Response(status=status2, text=text2)

            @MethodMatcher("*", resolver=self.resolver)
            async def other(request):
                return Response(status=status3, text=text3)

        async with run(AnotherApp(), self.resolver) as client:
            response = await client.get("/route")
            self.assertEqual(response.status, status2)
            self.assertEqual(await response.text(), text2)

            response2 = await client.get("/")
            self.assertEqual(response2.status, status3)
            self.assertEqual(await response2.text(), text3)

        async with run(App(), self.resolver) as client:
            response = await client.get("/other")
            self.assertEqual(response.status, status1)
            self.assertEqual(await response.text(), text1)

    @pytest.mark.asyncio
    async def test_app_multiple_inheritance(self):
        path1, status1, text1 = "/route-1", 201, "text-1"
        path2, status2, text2 = "/route-2", 202, "text-2"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path1, resolver=self.resolver)
            async def handler(request):
                return Response(status=status1, text=text1)

        class AnotherApp(jj.App):
            @PathMatcher(path2, resolver=self.resolver)
            async def handler(request):
                return Response(status=status2, text=text2)

        class ChildApp(App, AnotherApp):
            pass

        async with run(ChildApp(), self.resolver) as client:
            response = await client.get(path1)
            self.assertEqual(response.status, status1)
            self.assertEqual(await response.text(), text1)

            response2 = await client.get(path2)
            self.assertEqual(response2.status, status2)
            self.assertEqual(await response2.text(), text2)

    @pytest.mark.asyncio
    async def test_app_setter(self):
        path, status, text = "/route", 201, "text"

        @PathMatcher(path, resolver=self.resolver)
        async def handler(request):
            return Response(status=status, text=text)

        class App(jj.App):
            resolver = self.resolver
        App.handler = handler

        async with run(App(), self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, status)
            self.assertEqual(await response.text(), text)

        async with run(self.default_app, self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_app_deleter(self):
        path, status, text = "/route", 201, "text"

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher(path, resolver=resolver)
            async def handler(request):
                return Response(status=status, text=text)

        delattr(App, "handler")

        async with run(App(), self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_app_deleter_nonexisting_attr(self):
        class App(jj.App):
            resolver = self.resolver

        with self.assertRaises(AttributeError):
            delattr(App, "handler")

    @pytest.mark.asyncio
    async def test_app_instance_setter(self):
        path, status, text = "/route", 201, "text"

        @PathMatcher(path, resolver=self.resolver)
        async def handler(request):
            return Response(status=status, text=text)

        class App(jj.App):
            resolver = self.resolver
        app = App()
        app.handler = handler

        async with run(app, self.resolver) as client:
            response = await client.get(path)
            self.assertEqual(response.status, 404)



from typing import Any, Dict, List, Tuple, Union

from multidict import MultiDict, MultiMapping
from packed import packable

from ._attribute_matcher import AttributeMatcher
from ._equal_matcher import EqualMatcher

__all__ = ("MultiDictMatcher", "StrOrAttrMatcher", "DictOrTupleList",)


StrOrAttrMatcher = Union[str, AttributeMatcher]
DictOrTupleList = Union[
    Dict[str, StrOrAttrMatcher],
    List[Tuple[str, StrOrAttrMatcher]],
]


@packable("jj.matchers.MultiDictMatcher")
class MultiDictMatcher(AttributeMatcher):
    def __init__(self, expected: DictOrTupleList) -> None:
        self._expected = MultiDict(expected)

    async def _match_any(self, submatcher: AttributeMatcher, values: List[Any]) -> bool:
        for value in values:
            if await submatcher.match(value):
                return True
        return False

    async def match(self, actual: MultiMapping[str]) -> bool:
        assert isinstance(actual, MultiMapping)

        for key, val in self._expected.items():
            submatcher = val if isinstance(val, AttributeMatcher) else EqualMatcher(val)
            values: List[Any] = actual.getall(key, [])
            if not await self._match_any(submatcher, values):
                return False
        return True

    def __repr__(self) -> str:
        expected = [(key, val) for key, val in self._expected.items()]
        return f"{self.__class__.__qualname__}({expected!r})"

    def __packed__(self) -> Dict[str, Any]:
        expected = [[key, val] for key, val in self._expected.items()]
        return {"expected": expected}

    @classmethod
    def __unpacked__(cls, *, expected: DictOrTupleList, **kwargs: Any) -> "MultiDictMatcher":
        return cls(expected)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
    from unittest.mock import AsyncMock
else:
    from unittest import TestCase
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest

from jj.apps import create_app
from jj.resolvers import Registry, Resolver


class TestResolver(TestCase):
    def setUp(self):
        self.default_handler = AsyncMock(return_value=sentinel.default_response)
        self.default_app = create_app()
        self.resolver = Resolver(Registry(), self.default_app, self.default_handler)

    # Apps

    def test_get_apps(self):
        apps = self.resolver.get_apps()
        self.assertEqual(apps, [])

    def test_register_app(self):
        res = self.resolver.register_app(type(self.default_app))
        self.assertIsNone(res)

        apps = self.resolver.get_apps()
        self.assertEqual(apps, [type(self.default_app)])

    def test_register_another_app(self):
        self.resolver.register_app(type(self.default_app))

        app = create_app()
        self.resolver.register_app(type(app))

        apps = self.resolver.get_apps()
        self.assertEqual(apps, [type(self.default_app), type(app)])

    def test_register_app_twice(self):
        self.resolver.register_app(type(self.default_app))

        res = self.resolver.register_app(type(self.default_app))
        self.assertIsNone(res)

        apps = self.resolver.get_apps()
        self.assertEqual(apps, [type(self.default_app)])

    def test_deregister_single_app(self):
        self.resolver.register_app(type(self.default_app))

        res = self.resolver.deregister_app(type(self.default_app))
        self.assertIsNone(res)

        apps = self.resolver.get_apps()
        self.assertEqual(apps, [])

    def test_deregister_app(self):
        app1, app2 = create_app(), create_app()
        self.resolver.register_app(type(app1))
        self.resolver.register_app(type(app2))

        self.resolver.deregister_app(type(app1))

        apps = self.resolver.get_apps()
        self.assertEqual(apps, [type(app2)])

    def test_deregister_nonexisting_app(self):
        app = create_app()

        res = self.resolver.deregister_app(type(app))
        self.assertIsNone(res)

    # Handlers

    def test_get_handlers(self):
        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [])

    def test_get_handlers_with_nonexisting_app(self):
        app = create_app()

        handlers = self.resolver.get_handlers(type(app))
        self.assertEqual(handlers, [])

    def test_register_handler(self):
        handler = AsyncMock(return_value=sentinel.response)
        res = self.resolver.register_handler(handler, type(self.default_app))
        self.assertIsNone(res)

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler])

    def test_register_another_handler(self):
        handler1 = AsyncMock(return_value=sentinel.response)
        handler2 = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler1, type(self.default_app))

        self.resolver.register_handler(handler2, type(self.default_app))

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler1, handler2])

    def test_register_handler_twice(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        res = self.resolver.register_handler(handler, type(self.default_app))
        self.assertIsNone(res)

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler])

    def test_deregister_single_handler(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        res = self.resolver.deregister_handler(handler, type(self.default_app))
        self.assertIsNone(res)

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [])

    def test_deregister_handler(self):
        handler1 = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler1, type(self.default_app))
        handler2 = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler2, type(self.default_app))

        self.resolver.deregister_handler(handler1, type(self.default_app))

        handlers = self.resolver.get_handlers(type(self.default_app))
        self.assertEqual(handlers, [handler2])

    def test_deregister_nonexisting_handler(self):
        handler = AsyncMock(return_value=sentinel.response)

        res = self.resolver.deregister_handler(handler, type(self.default_app))
        self.assertIsNone(res)

    def test_deregister_handler_with_nonexisting_app(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        app = create_app()
        res = self.resolver.deregister_handler(handler, type(app))
        self.assertIsNone(res)

    # Attributes

    def test_get_nonexisting_attribute(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        attribute_value = self.resolver.get_attribute(sentinel.name, handler)
        self.assertEqual(attribute_value, sentinel)

    def test_get_attribute_with_non_existing_handler(self):
        handler = AsyncMock(return_value=sentinel.response)

        default = None
        attribute_value = self.resolver.get_attribute(sentinel.name, handler, default)
        self.assertEqual(attribute_value, default)

    def test_register_attribute(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        res = self.resolver.register_attribute(sentinel.name, sentinel.value, handler)
        self.assertIsNone(res)

        attribute_value = self.resolver.get_attribute(sentinel.name, handler)
        self.assertEqual(attribute_value, sentinel.value)

    def test_register_another_attribute(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))
        self.resolver.register_attribute(sentinel.name1, sentinel.value1, handler)

        self.resolver.register_attribute(sentinel.name2, sentinel.value2, handler)

        attribute_value2 = self.resolver.get_attribute(sentinel.name2, handler)
        self.assertEqual(attribute_value2, sentinel.value2)
        attribute_value1 = self.resolver.get_attribute(sentinel.name1, handler)
        self.assertEqual(attribute_value1, sentinel.value1)

    def test_register_attribute_twice(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))
        self.resolver.register_attribute(sentinel.name1, sentinel.value1, handler)

        res = self.resolver.register_attribute(sentinel.name, sentinel.value, handler)
        self.assertIsNone(res)

        attribute_value = self.resolver.get_attribute(sentinel.name, handler)
        self.assertEqual(attribute_value, sentinel.value)

    def test_deregister_attribute(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))
        self.resolver.register_attribute(sentinel.name, sentinel.value, handler)

        res = self.resolver.deregister_attribute(sentinel.name, handler)
        self.assertIsNone(res)

        attribute_value = self.resolver.get_attribute(sentinel.name, handler, default=None)
        self.assertEqual(attribute_value, None)

    def test_deregister_nonexisting_attribute(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        res = self.resolver.deregister_attribute(sentinel.name, handler)
        self.assertIsNone(res)

    def test_deregister_attribute_with_nonexisting_handler(self):
        handler = AsyncMock(return_value=sentinel.response)

        res = self.resolver.deregister_attribute(sentinel.name, handler)
        self.assertIsNone(res)

    # Matchers

    def test_get_matchers_without_matchers(self):
        matchers = self.resolver.get_matchers(self.default_handler)
        self.assertEqual(matchers, [])

    def test_get_matchers_with_one_matcher(self):
        matcher = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher, self.default_handler)

        matchers = self.resolver.get_matchers(self.default_handler)
        self.assertEqual(matchers, [matcher])

    def test_get_matchers_with_multiple_matchers(self):
        matcher1 = AsyncMock(return_value=True)
        matcher2 = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher1, self.default_handler)
        self.resolver.register_matcher(matcher2, self.default_handler)

        matchers = self.resolver.get_matchers(self.default_handler)
        self.assertEqual(matchers, [matcher1, matcher2])

    def test_register_matcher(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher = AsyncMock(return_value=True)
        self.assertIsNone(self.resolver.register_matcher(matcher, handler))

    def test_deregister_matcher(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher, handler)

        res = self.resolver.deregister_matcher(matcher, handler)
        self.assertIsNone(res)

        matchers = self.resolver.get_matchers(handler)
        self.assertEqual(matchers, [])

    def test_deregister_matcher_with_nonexisting_handler(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher = AsyncMock(return_value=True)

        res = self.resolver.deregister_matcher(matcher, handler)
        self.assertIsNone(res)

    # Resolver

    @pytest.mark.asyncio
    async def test_resolve_request_with_all_truthy_matchers(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher1 = AsyncMock(return_value=True)
        matcher2 = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher1, handler)
        self.resolver.register_matcher(matcher2, handler)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler)

        matcher1.assert_called_once_with(request)
        matcher2.assert_called_once_with(request)
        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_resolve_request_with_all_falsy_matchers(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher1 = AsyncMock(return_value=False)
        matcher2 = AsyncMock(return_value=False)
        self.resolver.register_matcher(matcher1, handler)
        self.resolver.register_matcher(matcher2, handler)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, self.default_handler)

        matcher1.assert_called_once_with(request)
        matcher2.assert_not_called()
        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_resolve_request_with_first_falsy_matcher(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher1 = AsyncMock(return_value=False)
        matcher2 = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher1, handler)
        self.resolver.register_matcher(matcher2, handler)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, self.default_handler)

        matcher1.assert_called_once_with(request)
        matcher2.assert_not_called()
        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_resolve_request_with_last_falsy_matcher(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher1 = AsyncMock(return_value=True)
        matcher2 = AsyncMock(return_value=False)
        self.resolver.register_matcher(matcher1, handler)
        self.resolver.register_matcher(matcher2, handler)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, self.default_handler)

        matcher1.assert_called_once_with(request)
        matcher2.assert_called_once_with(request)
        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_resolve_request_without_handlers(self):
        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, self.default_handler)

    @pytest.mark.asyncio
    async def test_resolve_request_with_nonexisting_app(self):
        app = create_app()
        request = Mock()
        response = await self.resolver.resolve(request, app)
        self.assertEqual(response, self.default_handler)

    @pytest.mark.asyncio
    async def test_resolve_request_without_matchers(self):
        handler = AsyncMock(return_value=sentinel.response)
        self.resolver.register_handler(handler, type(self.default_app))

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, self.default_handler)

    @pytest.mark.asyncio
    async def test_resolve_request_with_single_matcher(self):
        handler = AsyncMock(return_value=sentinel.response)
        matcher = AsyncMock(return_value=True)
        self.resolver.register_matcher(matcher, handler)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler)

        matcher.assert_called_once_with(request)
        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_resolve_request_with_multiple_handlers(self):
        matcher = AsyncMock(side_effect=(False, True))
        handler1 = AsyncMock(return_value=sentinel.response1)
        handler2 = AsyncMock(return_value=sentinel.response2)
        self.resolver.register_matcher(matcher, handler1)
        self.resolver.register_matcher(matcher, handler2)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler1)

        handler1.assert_not_called()
        handler2.assert_not_called()
        matcher.assert_has_calls([call(request)] * 2, any_order=True)
        self.assertEqual(matcher.call_count, 2)

    @pytest.mark.asyncio
    async def test_resolve_request_priority(self):
        matcher = AsyncMock(side_effect=(True, True))
        handler1 = AsyncMock(return_value=sentinel.response1)
        handler2 = AsyncMock(return_value=sentinel.response2)
        self.resolver.register_matcher(matcher, handler1)
        self.resolver.register_matcher(matcher, handler2)

        request = Mock()
        response = await self.resolver.resolve(request, self.default_app)
        self.assertEqual(response, handler2)

        handler1.assert_not_called()
        handler2.assert_not_called()
        matcher.assert_called_once_with(request)



from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("ExistMatcher",)


@packable("jj.matchers.ExistMatcher")
class ExistMatcher(AttributeMatcher):
    async def match(self, actual: Any) -> bool:
        return True

    def __packed__(self) -> Dict[str, Any]:
        return {}

    @classmethod
    def __unpacked__(cls, **kwargs: Any) -> "ExistMatcher":
        return cls()



from typing import Any, Dict, Union

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, DictOrTupleList, MultiDictMatcher
from ._request_matcher import RequestMatcher

__all__ = ("HeaderMatcher", "DictOrTupleListOrAttrMatcher",)

DictOrTupleListOrAttrMatcher = Union[DictOrTupleList, AttributeMatcher]


@packable("jj.matchers.HeaderMatcher")
class HeaderMatcher(RequestMatcher):
    def __init__(self, headers: DictOrTupleListOrAttrMatcher, *, resolver: Resolver) -> None:
        super().__init__(resolver=resolver)
        if isinstance(headers, AttributeMatcher):
            self._matcher = headers
        else:
            self._matcher = MultiDictMatcher(headers)

    async def match(self, request: Request) -> bool:
        return await self._matcher.match(request.headers)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        return {"headers": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     headers: DictOrTupleListOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "HeaderMatcher":
        return cls(headers, resolver=resolver)



from unittest.mock import Mock, call, sentinel

import pytest

from jj.matchers import AttributeMatcher, ContainMatcher, NotContainMatcher

from ..._test_utils.steps import given, then, when


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ("1", ["1"], True),
    ("1", ["2", "1"], True),
    ("2", ["1"], False),
    ("2", ["1", "3"], False),
])
async def test_contain_matcher(expected, actual, res):
    with given:
        matcher = ContainMatcher(expected)

    with when:
        actual = await matcher.match(actual)

    with then:
        assert actual is res


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ("1", ["1"], False),
    ("1", ["2", "1"], False),
    ("2", ["1"], True),
    ("2", ["1", "3"], True),
])
async def test_not_not_contain_matcher(expected, actual, res):
    with given:
        matcher = NotContainMatcher(expected)

    with when:
        actual = await matcher.match(actual)

    with then:
        assert actual is res


@pytest.mark.asyncio
@pytest.mark.parametrize("ret_val", [True, False])
async def test_contain_matcher_with_magic_method(ret_val):
    with given:
        rec_ = Mock(__contains__=Mock(return_value=ret_val))
        matcher = ContainMatcher(sentinel.expected)

    with when:
        actual = await matcher.match(rec_)

    with then:
        assert actual is ret_val
        assert rec_.__contains__.mock_calls == [call(sentinel.expected)]


@pytest.mark.asyncio
@pytest.mark.parametrize("ret_val", [True, False])
async def test_not_contain_matcher_with_magic_method(ret_val):
    with given:
        rec_ = Mock(__contains__=Mock(return_value=not ret_val))
        matcher = NotContainMatcher(sentinel.expected)

    with when:
        actual = await matcher.match(rec_)

    with then:
        assert actual is ret_val
        assert rec_.__contains__.mock_calls == [call(sentinel.expected)]


@pytest.mark.parametrize("matcher_class", [ContainMatcher, NotContainMatcher])
def test_is_instance_of_attribute_matcher(matcher_class):
    with given:
        matcher = matcher_class(sentinel.expected)

    with when:
        actual = isinstance(matcher, AttributeMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize(("expected", "matcher_class"), [
    ("ContainMatcher('smth')", ContainMatcher),
    ("NotContainMatcher('smth')", NotContainMatcher),
])
def test_repr(expected, matcher_class):
    with given:
        matcher = matcher_class("smth")

    with when:
        actual = repr(matcher)

    with then:
        assert actual == expected


@pytest.mark.parametrize("matcher_class", [ContainMatcher, NotContainMatcher])
def test_pack(matcher_class):
    with given:
        expected = "smth"
        matcher = matcher_class(expected)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"expected": expected}


@pytest.mark.parametrize("matcher_class", [ContainMatcher, NotContainMatcher])
def test_unpack(matcher_class):
    with given:
        kwargs = {
            "expected": "smth",
            "future_field": sentinel,
        }

    with when:
        actual = matcher_class.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, matcher_class)



from unittest.mock import sentinel

import pytest

from jj.matchers import AttributeMatcher, ExistMatcher

from ..._test_utils.steps import given, then, when


@pytest.mark.asyncio
@pytest.mark.parametrize("value", [None, "smth", sentinel.value])
async def test_exist_matcher(value):
    with given:
        matcher = ExistMatcher()

    with when:
        actual = await matcher.match(value)

    with then:
        assert actual is True


def test_is_instance_of_attribute_matcher():
    with given:
        matcher = ExistMatcher()

    with when:
        actual = isinstance(matcher, AttributeMatcher)

    with then:
        assert actual is True


def test_repr():
    with given:
        matcher = ExistMatcher()

    with when:
        actual = repr(matcher)

    with then:
        assert actual == "ExistMatcher()"


def test_pack():
    with given:
        matcher = ExistMatcher()

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {}


def test_unpack():
    with given:
        kwargs = {"future_field": sentinel}

    with when:
        actual = ExistMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, ExistMatcher)



import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

from unittest.mock import Mock

import pytest

import jj
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher, PathMatcher
from jj.middlewares import BaseMiddleware, RootMiddleware, SelfMiddleware
from jj.requests import Request
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestSelfMiddleware(TestCase):
    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    def test_middleware_without_resolver(self):
        with self.assertRaises(Exception):
            SelfMiddleware()

    def test_middleware_with_instance_resolver(self):
        middleware = SelfMiddleware(self.resolver)
        self.assertIsInstance(middleware, RootMiddleware)

    def test_middleware_with_class_resolver(self):
        class Middleware(SelfMiddleware):
            resolver = self.resolver
        middleware = Middleware()
        self.assertIsInstance(middleware, RootMiddleware)

    @pytest.mark.asyncio
    async def test_middleware_without_handlers(self):
        class App(jj.App):
            resolver = self.resolver

        middleware = SelfMiddleware(self.resolver)
        async with run(App(), middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 404)

    @pytest.mark.asyncio
    async def test_middleware_with_handler(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                mock(request)
                return Response(status=200)

        middleware = SelfMiddleware(self.resolver)
        async with run(App(), middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_called_once()
        request_arg, = mock.call_args[0]
        self.assertIsInstance(request_arg, Request)

    @pytest.mark.asyncio
    async def test_middleware_with_handler_and_self_reference(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(self, request):
                mock(self, request)
                return Response(status=200)

        app, middleware = App(), SelfMiddleware(self.resolver)
        async with run(app, middlewares=[middleware]) as client:
            response = await client.get("/")
            self.assertEqual(response.status, 200)

        mock.assert_called_once()
        app_arg, request_arg = mock.call_args[0]
        self.assertIsInstance(app_arg, App)
        self.assertIsInstance(request_arg, Request)

    @pytest.mark.asyncio
    async def test_middleware_with_handler_middleware(self):
        mock = Mock()

        class App(jj.App):
            resolver = self.resolver

            @PathMatcher("/path", resolver=resolver)
            @BaseMiddleware(resolver)
            @MethodMatcher("*", resolver=resolver)
            async def handler(self, request):
                mock(self, request)
                return Response(status=200)

        app, middleware = App(), SelfMiddleware(self.resolver)
        async with run(app, middlewares=[middleware]) as client:
            response = await client.get("/path")
            self.assertEqual(response.status, 200)

        mock.assert_called_once()
        app_arg, request_arg = mock.call_args[0]
        self.assertIsInstance(app_arg, App)
        self.assertIsInstance(request_arg, Request)



from typing import Any, Dict, List, Tuple

from multidict import CIMultiDict, CIMultiDictProxy
from packed import packable

from ...responses import Response, StreamResponse

__all__ = ("HistoryResponse",)


@packable("jj.mock.HistoryResponse")
class HistoryResponse:
    def __init__(self, *,
                 status: int,
                 reason: str,
                 headers: "CIMultiDictProxy[str]",
                 body: Any,
                 raw: bytes) -> None:
        self._status = status
        self._reason = reason
        self._headers = headers
        self._body = body
        self._raw = raw

    @property
    def status(self) -> int:
        return self._status

    @property
    def reason(self) -> str:
        return self._reason

    @property
    def headers(self) -> "CIMultiDictProxy[str]":
        return self._headers

    @property
    def body(self) -> Any:
        return self._body

    @property
    def raw(self) -> bytes:
        return self._raw

    @classmethod
    async def from_response(cls, response: StreamResponse) -> "HistoryResponse":
        if isinstance(response, Response):
            raw = response.get_body()
        else:
            raw = b"<stream response>"
        return cls(
            status=response.status,
            reason=response.reason,
            headers=CIMultiDictProxy(response.headers),
            body=raw,
            raw=raw,
        )

    def to_dict(self) -> Dict[str, Any]:
        headers = [[key, val] for key, val in self._headers.items()]
        return {
            "status": self._status,
            "reason": self._reason,
            "headers": headers,
            "body": self._body,
            "raw": self._raw,
        }

    def __packed__(self) -> Dict[str, Any]:
        return self.to_dict()

    @classmethod
    def from_dict(cls, request: Dict[str, Any]) -> "HistoryResponse":
        real_headers = CIMultiDictProxy(CIMultiDict(request["headers"]))
        raw = request.get("raw", request["body"])  # backward compatibility
        return cls(
            status=request["status"],
            reason=request["reason"],
            headers=real_headers,
            body=request["body"],
            raw=raw,
        )

    @classmethod
    def __unpacked__(cls, *,
                     status: int,
                     reason: str,
                     headers: List[Tuple[str, str]],
                     body: Any,
                     **kwargs: Any) -> "HistoryResponse":
        return cls.from_dict({
            "status": status,
            "reason": reason,
            "headers": headers,
            "body": body,
            **kwargs,
        })

    def __repr__(self) -> str:
        return (f"HistoryResponse("
                f"status={self._status!r}, "
                f"reason={self._reason!r}, "
                f"headers={self._headers!r}, "
                f"body={self._body!r}"
                f")")



import asyncio

from aiohttp.test_utils import TestClient

from .server import TestServer

__all__ = ("TestClient", "TestServer", "run",)


def run(app, resolver=None, middlewares=None):
    loop = asyncio.get_event_loop()
    return TestClient(TestServer(app, resolver, middlewares, loop=loop), loop=loop)



from jj.expiration_policy import ExpireAfterRequests, ExpireNever

from .._test_utils.steps import given, then, when


def test_pack_expire_never():
    with given:
        expiration_policy = ExpireNever()

    with when:
        actual = expiration_policy.__packed__()

    with then:
        assert actual == {}


def test_pack_expire_after_requests():
    with given:
        expiration_policy = ExpireAfterRequests(2)

    with when:
        actual = expiration_policy.__packed__()

    with then:
        assert actual == {
            "max_requests_count": 2,
        }



from typing import Any

__all__ = ("AttributeMatcher",)


class AttributeMatcher:
    async def match(self, actual: Any) -> bool:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



import os
import sys

if sys.version_info >= (3, 8):
    from unittest import IsolatedAsyncioTestCase as TestCase
else:
    from unittest import TestCase

import pytest

import jj
from jj import server_version
from jj.apps import create_app
from jj.handlers import default_handler
from jj.matchers import MethodMatcher
from jj.resolvers import Registry, ReversedResolver
from jj.responses import Response

from .._test_utils import run


class TestResponse(TestCase):
    def make_app_with_response(self, *args, **kwargs):
        class App(jj.App):
            resolver = self.resolver

            @MethodMatcher("*", resolver=resolver)
            async def handler(request):
                return Response(*args, **kwargs)
        return App()

    def make_path(self, path):
        return os.path.join(os.path.dirname(__file__), path)

    def setUp(self):
        self.default_app = create_app()
        self.resolver = ReversedResolver(Registry(), self.default_app, default_handler)

    @pytest.mark.asyncio
    async def test_response_with_default_args(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.get("/")
            # status
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, "OK")
            # headers
            self.assertEqual(response.headers.get("Server"), server_version)
            self.assertEqual(response.headers.get("Content-Length"), "0")
            self.assertEqual(response.headers.get("Content-Type"), "text/plain; charset=utf-8")
            self.assertIsNotNone(response.headers.get("Date"))
            self.assertEqual(len(response.headers), 4)
            # body
            raw = await response.read()
            self.assertEqual(raw, b"")

    @pytest.mark.asyncio
    async def test_response_with_conflicted_args(self):
        payload = "200 OK"

        with self.assertRaises(Exception):
            Response(text=payload, body=payload)

        with self.assertRaises(Exception):
            Response(text=payload, json=payload)

        with self.assertRaises(Exception):
            Response(body=payload, json=payload)

    # Status

    @pytest.mark.asyncio
    async def test_response_status(self):
        status = 204
        app = self.make_app_with_response(status=status)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, status)
            # aiohttp автоматически подставляет нужный reason
            self.assertEqual(response.reason, "No Content")

    @pytest.mark.asyncio
    async def test_response_reason(self):
        status, reason = 204, "Custom Reason"
        app = self.make_app_with_response(status=status, reason=reason)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.status, status)
            self.assertEqual(response.reason, reason)

    # Body

    @pytest.mark.asyncio
    async def test_response_body(self):
        body = "200 OK"
        binary_body = b"200 OK"
        app = self.make_app_with_response(body=body)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), body)
            self.assertEqual(await response.read(), binary_body)

            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "text/plain; charset=utf-8")

    @pytest.mark.asyncio
    async def test_response_text_body(self):
        text = "200 OK"
        binary_text = b"200 OK"
        app = self.make_app_with_response(text=text)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), text)
            self.assertEqual(await response.read(), binary_text)

            self.assertEqual(response.headers.get("Content-Length"), str(len(text)))
            self.assertEqual(response.headers.get("Content-Type"), "text/plain; charset=utf-8")

    @pytest.mark.asyncio
    async def test_response_json_body(self):
        json = {"key": None}
        dumped_json = '{"key": null}'
        binary_json = b'{"key": null}'
        app = self.make_app_with_response(json=json)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.json(), json)
            self.assertEqual(await response.text(), dumped_json)
            self.assertEqual(await response.read(), binary_json)

            self.assertEqual(response.headers.get("Content-Length"), str(len(dumped_json)))
            self.assertEqual(response.headers.get("Content-Type"), "application/json")

    @pytest.mark.asyncio
    async def test_response_json_body_with_custom_content_type(self):
        json = {"key": None}
        dumped_json = '{"key": null}'
        binary_json = b'{"key": null}'
        content_type = "text/plain"
        app = self.make_app_with_response(json=json, headers={"Content-Type": content_type})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), dumped_json)
            self.assertEqual(await response.read(), binary_json)

            self.assertEqual(response.headers.get("Content-Length"), str(len(dumped_json)))
            self.assertEqual(response.headers.get("Content-Type"), content_type)

    @pytest.mark.asyncio
    async def test_response_null_json_body(self):
        json = None
        dumped_json = 'null'
        binary_json = b'null'
        app = self.make_app_with_response(json=json)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.json(), json)
            self.assertEqual(await response.text(), dumped_json)
            self.assertEqual(await response.read(), binary_json)

            self.assertEqual(response.headers.get("Content-Length"), str(len(dumped_json)))
            self.assertEqual(response.headers.get("Content-Type"), "application/json")

    @pytest.mark.asyncio
    async def test_response_binary_body(self):
        body = b"200 OK"
        app = self.make_app_with_response(body=body)

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), body.decode())
            self.assertEqual(await response.read(), body)

            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "application/octet-stream")

    @pytest.mark.asyncio
    async def test_response_predefined_text_body(self):
        path = self.make_path("fixtures/users.json")
        with open(path, "rt") as f:
            body = f.read()
        app = self.make_app_with_response(body=open(path, "rt"))

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), body)
            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "text/plain; charset=utf-8")
            self.assertEqual(response.headers.get("Content-Disposition"), "inline")

    @pytest.mark.asyncio
    async def test_response_predefined_binary_body(self):
        path = self.make_path("fixtures/users.json")
        with open(path, "rb") as f:
            body = f.read()
        app = self.make_app_with_response(body=open(path, "rb"))

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(await response.text(), body.decode())
            self.assertEqual(await response.read(), body)
            self.assertEqual(response.headers.get("Content-Length"), str(len(body)))
            self.assertEqual(response.headers.get("Content-Type"), "application/json")
            self.assertEqual(response.headers.get("Content-Disposition"), "inline")

    # Headers

    @pytest.mark.asyncio
    async def test_response_header(self):
        custom_header_key, custom_header_val = "Cutom-Header", "Value"
        app = self.make_app_with_response(headers={custom_header_key: custom_header_val})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get(custom_header_key), custom_header_val)
            self.assertEqual(response.headers.get("Server"), server_version)

    @pytest.mark.asyncio
    async def test_response_headers(self):
        custom_header1_key, custom_header1_val = "Cutom-Header", "Value1"
        custom_header2_key, custom_header2_val = "Cutom-Header", "Value2"
        app = self.make_app_with_response(headers=[
            (custom_header1_key, custom_header1_val),
            (custom_header2_key, custom_header2_val),
        ])

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.getall(custom_header1_key),
                             [custom_header1_val, custom_header2_val])
            self.assertEqual(response.headers.get("Server"), server_version)

    @pytest.mark.asyncio
    async def test_response_with_custom_server_header(self):
        server_header = "server version x"
        app = self.make_app_with_response(headers={"Server": server_header})

        async with run(app) as client:
            response = await client.get("/")
            self.assertEqual(response.headers.get("Server"), server_header)

    @pytest.mark.asyncio
    async def test_response_with_expect_header(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.post("/", json={}, expect100=True)
            self.assertEqual(response.status, 200)

    @pytest.mark.asyncio
    async def test_response_with_incorrect_expect_header(self):
        app = self.make_app_with_response()

        async with run(app) as client:
            response = await client.post("/", json={}, headers={"Expect": "banana"})
            self.assertEqual(response.status, 417)



import pytest
from aiohttp.web import ContentCoding

from jj.responses import Response

from .._test_utils.steps import given, then, when


def eq(r1, r2):
    assert r1.status == r2.status
    assert r1.reason == r2.reason
    assert r1.body == r2.body
    assert r1.headers == r2.headers
    assert r1.cookies == r2.cookies
    assert r1.chunked == r2.chunked
    assert r1.compression == r2.compression
    assert r1.content_coding == r2.content_coding


@pytest.mark.parametrize(("attr", "val"), [
    ("status", 204),
    ("reason", "not found"),
    ("body", b"text"),
    ("headers", [
        ("key", "1"),
        ("key", "2"),
        ("another_key", "3"),
    ]),
])
def test_response_attrs(attr, val):
    with given:
        response = Response(**{attr: val})

    with when:
        copy = response.copy()

    with then:
        assert eq(copy, response) is None


def test_response_cookies():
    with given:
        response = Response()
        response.set_cookie("name", "1")
        response.set_cookie("another_name", "2")

    with when:
        copy = response.copy()

    with then:
        assert eq(copy, response) is None


def test_response_chunked():
    with given:
        response = Response(body=b"text")
        response.enable_chunked_encoding()

    with when:
        copy = response.copy()

    with then:
        assert eq(copy, response) is None


def test_response_compression():
    with given:
        response = Response(body=b"text")
        response.enable_compression(ContentCoding.gzip)

    with when:
        copy = response.copy()

    with then:
        assert eq(copy, response) is None



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock, call, sentinel

import pytest

from jj.matchers import AttributeMatcher, MethodMatcher, RequestMatcher
from jj.matchers.attribute_matchers import EqualMatcher

from ..._test_utils.fixtures import request_, resolver_
from ..._test_utils.steps import given, then, when

__all__ = ("request_", "resolver_",)


@pytest.mark.asyncio
@pytest.mark.parametrize(("expected", "actual", "res"), [
    ("*", "GET", True),
    ("*", "POST", True),
    ("*", "CUSTOM", True),

    ("GET", "GET", True),
    ("get", "GET", True),
    ("POST", "POST", True),
    ("GET", "POST", False),
])
async def test_method_matcher(expected, actual, res, *, resolver_, request_):
    with given:
        request_.method = actual
        matcher = MethodMatcher(expected, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res


@pytest.mark.asyncio
@pytest.mark.parametrize(("ret_vals", "res", "called_with"), [
    ([True, True], True, ["*"]),
    ([True, False], True, ["*"]),
    ([False, True], True, ["*", "GET"]),
    ([False, False], False, ["*", "GET"]),
])
async def test_method_matcher_with_custom_submatcher(ret_vals, res, called_with, *,
                                                     resolver_, request_):
    with given:
        request_.method = "GET"
        submatcher_ = Mock(AttributeMatcher, match=AsyncMock(side_effect=ret_vals))
        matcher = MethodMatcher(submatcher_, resolver=resolver_)

    with when:
        actual = await matcher.match(request_)

    with then:
        assert actual is res
        assert submatcher_.mock_calls == [call.match(x) for x in called_with]


def test_is_instance_of_request_matcher(*, resolver_):
    with given:
        matcher = MethodMatcher("*", resolver=resolver_)

    with when:
        actual = isinstance(matcher, RequestMatcher)

    with then:
        assert actual is True


@pytest.mark.parametrize(("method", "representation"), [
    ("*", "*"),
    ("GET", "GET"),
    ("post", "POST"),
])
def test_repr(method, representation, *, resolver_):
    with given:
        resolver_.__repr__ = Mock(return_value="<Resolver>")
        matcher = MethodMatcher(method, resolver=resolver_)

    with when:
        actual = repr(matcher)

    with then:
        assert actual == f"MethodMatcher(EqualMatcher({representation!r}), resolver=<Resolver>)"


def test_pack():
    with given:
        submatcher = EqualMatcher("GET")
        matcher = MethodMatcher(submatcher, resolver=resolver_)

    with when:
        actual = matcher.__packed__()

    with then:
        assert actual == {"method": submatcher}


def test_unpack():
    with given:
        submatcher = EqualMatcher("GET")
        kwargs = {
            "method": submatcher,
            "resolver": resolver_,
            "future_field": sentinel,
        }

    with when:
        actual = MethodMatcher.__unpacked__(**kwargs)

    with then:
        assert isinstance(actual, MethodMatcher)



from asyncio import AbstractEventLoop, ensure_future
from typing import Any, List, Optional, Type

from aiohttp.web_runner import BaseRunner, TCPSite

__all__ = ("Server",)


class Server:
    def __init__(self, loop: AbstractEventLoop,
                 runner_factory: Type[BaseRunner],
                 site_factory: Type[TCPSite]) -> None:
        self._loop = loop
        self._runner_factory = runner_factory
        self._site_factory = site_factory
        self._runners: List[BaseRunner] = []

    async def _start(self, app: Any, host: Optional[str], port: Optional[int]) -> None:
        runner = self._runner_factory(app, loop=self._loop)  # type: ignore
        await runner.setup()

        site = self._site_factory(runner, host=host, port=port)
        await site.start()

        self._runners += [runner]

    def start(self, app: Any, host: Optional[str] = None, port: Optional[int] = None) -> None:
        ensure_future(self._start(app, host, port))

    def serve(self) -> None:
        self._loop.run_forever()

    def cleanup(self) -> None:
        for runner in reversed(self._runners):
            self._loop.run_until_complete(runner.cleanup())

    def shutdown(self) -> None:
        self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        self._loop.close()



from aiohttp.test_utils import TestServer as BaseTestServer

from jj.runners import AppRunner


class TestServer(BaseTestServer):
    def __init__(self, app, resolver=None, middlewares=None, **kwargs):
        super().__init__(app, **kwargs)
        self.app = app
        self.resolver = resolver
        self.middlewares = middlewares or []

    async def _make_runner(self, **kwargs):
        try:
            resolver = self.app.resolver
        except (AttributeError, NotImplementedError):
            resolver = self.resolver
        return AppRunner(self.app, resolver, self.middlewares, self._loop)
